"""This script generates devinfo files for the test suite.

If you have new, yet unsupported device or a device with no devinfo file under
 tests/fixtures, feel free to run this script and create a PR to add the file
 to the repository.

Executing this script will several modules and methods one by one,
and finally execute a query to query all of them at once.
"""

from __future__ import annotations

import dataclasses
import json
import logging
import re
import sys
import traceback
from collections import defaultdict, namedtuple
from collections.abc import Callable
from pathlib import Path
from pprint import pprint
from typing import Any

import asyncclick as click

from devtools.helpers.smartcamrequests import SMARTCAM_REQUESTS
from devtools.helpers.smartrequests import SmartRequest, get_component_requests
from kasa import (
    AuthenticationError,
    Credentials,
    Device,
    DeviceConfig,
    DeviceConnectionParameters,
    Discover,
    KasaException,
    TimeoutError,
)
from kasa.device_factory import get_protocol
from kasa.deviceconfig import DeviceEncryptionType, DeviceFamily
from kasa.discover import (
    NEW_DISCOVERY_REDACTORS,
    DiscoveredRaw,
    DiscoveryResult,
)
from kasa.exceptions import SmartErrorCode
from kasa.protocols import IotProtocol
from kasa.protocols.iotprotocol import REDACTORS as IOT_REDACTORS
from kasa.protocols.protocol import redact_data
from kasa.protocols.smartcamprotocol import (
    SmartCamProtocol,
    _ChildCameraProtocolWrapper,
)
from kasa.protocols.smartprotocol import REDACTORS as SMART_REDACTORS
from kasa.protocols.smartprotocol import SmartProtocol, _ChildProtocolWrapper
from kasa.smart import SmartChildDevice, SmartDevice
from kasa.smartcam import SmartCamChild, SmartCamDevice
from kasa.smartcam.smartcamchild import CHILD_INFO_FROM_PARENT

Call = namedtuple("Call", "module method")
FixtureResult = namedtuple("FixtureResult", "filename, folder, data, protocol_suffix")

SMART_FOLDER = "tests/fixtures/smart/"
SMARTCAM_FOLDER = "tests/fixtures/smartcam/"
SMART_CHILD_FOLDER = "tests/fixtures/smart/child/"
SMARTCAM_CHILD_FOLDER = "tests/fixtures/smartcam/child/"
IOT_FOLDER = "tests/fixtures/iot/"

SMART_PROTOCOL_SUFFIX = "SMART"
SMARTCAM_SUFFIX = "SMARTCAM"
SMART_CHILD_SUFFIX = "SMART.CHILD"
SMARTCAM_CHILD_SUFFIX = "SMARTCAM.CHILD"
IOT_SUFFIX = "IOT"

NO_GIT_FIXTURE_FOLDER = "kasa-fixtures"

ENCRYPT_TYPES = [encrypt_type.value for encrypt_type in DeviceEncryptionType]

_LOGGER = logging.getLogger(__name__)


def _wrap_redactors(redactors: dict[str, Callable[[Any], Any] | None]):
    """Wrap the redactors for  dump_devinfo.

    Will replace all partial REDACT_ values with zeros.
    If the data item is already scrubbed by dump_devinfo will leave as-is.
    """

    def _wrap(key: str) -> Any:
        def _wrapped(redactor: Callable[[Any], Any] | None) -> Any | None:
            if redactor is None:
                return lambda x: "**SCRUBBED**"

            def _redact_to_zeros(x: Any) -> Any:
                if isinstance(x, str) and "REDACT" in x:
                    return re.sub(r"\w", "0", x)
                if isinstance(x, dict):
                    for k, v in x.items():
                        x[k] = _redact_to_zeros(v)
                return x

            def _scrub(x: Any) -> Any:
                if key in {"ip", "local_ip"}:
                    return "127.0.0.123"
                # Already scrubbed by dump_devinfo
                if isinstance(x, str) and "SCRUBBED" in x:
                    return x
                default = redactor(x)
                return _redact_to_zeros(default)

            return _scrub

        return _wrapped(redactors[key])

    return {key: _wrap(key) for key in redactors}


@dataclasses.dataclass
class SmartCall:
    """Class for smart and smartcam calls."""

    module: str
    request: dict
    should_succeed: bool
    child_device_id: str
    supports_multiple: bool = True


def default_to_regular(d):
    """Convert nested defaultdicts to regular ones.

    From https://stackoverflow.com/a/26496899
    """
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d


async def handle_device(
    basedir, autosave, protocol, *, discovery_info=None, batch_size: int
):
    """Create a fixture for a single device instance."""
    if isinstance(protocol, SmartProtocol):
        fixture_results: list[FixtureResult] = await get_smart_fixtures(
            protocol, discovery_info=discovery_info, batch_size=batch_size
        )
    else:
        fixture_results = [
            await get_legacy_fixture(protocol, discovery_info=discovery_info)
        ]

    for fixture_result in fixture_results:
        save_folder = Path(basedir) / fixture_result.folder
        if save_folder.exists():
            save_filename = save_folder / f"{fixture_result.filename}.json"
        else:
            # If being run without git clone
            save_folder = Path(basedir) / NO_GIT_FIXTURE_FOLDER
            save_folder.mkdir(exist_ok=True)
            save_filename = (
                save_folder
                / f"{fixture_result.filename}-{fixture_result.protocol_suffix}.json"
            )

        pprint(fixture_result.data)
        if autosave:
            save = "y"
        else:
            save = click.prompt(
                f"Do you want to save the above content to {save_filename} (y/n)"
            )
        if save == "y":
            click.echo(f"Saving info to {save_filename}")

            with save_filename.open("w") as f:
                json.dump(fixture_result.data, f, sort_keys=True, indent=4)
                f.write("\n")
        else:
            click.echo("Not saving.")


@click.command()
@click.option("--host", required=False, help="Target host.")
@click.option(
    "--target",
    required=False,
    default="255.255.255.255",
    help="Target network for discovery.",
)
@click.option(
    "--username",
    default="",
    required=False,
    envvar="KASA_USERNAME",
    help="Username/email address to authenticate to device.",
)
@click.option(
    "--password",
    default="",
    required=False,
    envvar="KASA_PASSWORD",
    help="Password to use to authenticate to device.",
)
@click.option("--basedir", help="Base directory for the git repository", default=".")
@click.option("--autosave", is_flag=True, default=False, help="Save without prompting")
@click.option(
    "--batch-size", default=5, help="Number of batched requests to send at once"
)
@click.option("-d", "--debug", is_flag=True)
@click.option(
    "-di",
    "--discovery-info",
    help=(
        "Bypass discovery by passing an accurate discovery result json escaped string."
        + " Do not use this flag unless you are sure you know what it means."
    ),
)
@click.option(
    "--discovery-timeout",
    envvar="KASA_DISCOVERY_TIMEOUT",
    default=10,
    required=False,
    show_default=True,
    help="Timeout for discovery.",
)
@click.option(
    "-e",
    "--encrypt-type",
    envvar="KASA_ENCRYPT_TYPE",
    default=None,
    type=click.Choice(ENCRYPT_TYPES, case_sensitive=False),
)
@click.option(
    "-df",
    "--device-family",
    envvar="KASA_DEVICE_FAMILY",
    default="SMART.TAPOPLUG",
    help="Device family type, e.g. `SMART.KASASWITCH`.",
)
@click.option(
    "-lv",
    "--login-version",
    envvar="KASA_LOGIN_VERSION",
    default=2,
    type=int,
    help="The login version for device authentication. Defaults to 2",
)
@click.option(
    "--https/--no-https",
    envvar="KASA_HTTPS",
    default=False,
    is_flag=True,
    type=bool,
    help="Set flag if the device encryption uses https.",
)
@click.option(
    "--timeout",
    required=False,
    default=15,
    help="Timeout for queries.",
)
@click.option("--port", help="Port override", type=int)
async def cli(
    host,
    target,
    basedir,
    autosave,
    debug,
    username,
    discovery_timeout,
    password,
    batch_size,
    discovery_info,
    encrypt_type,
    https,
    device_family,
    login_version,
    port,
    timeout,
):
    """Generate devinfo files for devices.

    Use --host (for a single device) or --target (for a complete network).
    """
    if debug:
        logging.basicConfig(level=logging.DEBUG)

    raw_discovery = {}

    def capture_raw(discovered: DiscoveredRaw):
        raw_discovery[discovered["meta"]["ip"]] = discovered["discovery_response"]

    credentials = Credentials(username=username, password=password)
    if host is not None:
        if discovery_info:
            click.echo(f"Host and discovery info given, trying connect on {host}.")

            di = json.loads(discovery_info)
            dr = DiscoveryResult.from_dict(di)
            connection_type = DeviceConnectionParameters.from_values(
                dr.device_type,
                dr.mgt_encrypt_schm.encrypt_type,
                login_version=dr.mgt_encrypt_schm.lv,
                https=dr.mgt_encrypt_schm.is_support_https,
                http_port=dr.mgt_encrypt_schm.http_port,
            )
            dc = DeviceConfig(
                host=host,
                connection_type=connection_type,
                port_override=port,
                credentials=credentials,
                timeout=timeout,
            )
            device = await Device.connect(config=dc)
            await handle_device(
                basedir,
                autosave,
                device.protocol,
                discovery_info=dr.to_dict(),
                batch_size=batch_size,
            )
        elif device_family and encrypt_type:
            ctype = DeviceConnectionParameters(
                DeviceFamily(device_family),
                DeviceEncryptionType(encrypt_type),
                login_version,
                https,
            )
            config = DeviceConfig(
                host=host,
                port_override=port,
                credentials=credentials,
                connection_type=ctype,
                timeout=timeout,
            )
            if protocol := get_protocol(config):
                await handle_device(basedir, autosave, protocol, batch_size=batch_size)
            else:
                raise KasaException(
                    "Could not find a protocol for the given parameters."
                )
        else:
            click.echo(f"Host given, performing discovery on {host}.")
            device = await Discover.discover_single(
                host,
                credentials=credentials,
                port=port,
                discovery_timeout=discovery_timeout,
                timeout=timeout,
                on_discovered_raw=capture_raw,
            )
            discovery_info = raw_discovery[device.host]
            if decrypted_data := device._discovery_info.get("decrypted_data"):
                discovery_info["result"]["decrypted_data"] = decrypted_data
            await handle_device(
                basedir,
                autosave,
                device.protocol,
                discovery_info=discovery_info,
                batch_size=batch_size,
            )
    else:
        click.echo(
            "No --host given, performing discovery on"
            f" {target}. Use --target to override."
        )
        devices = await Discover.discover(
            target=target,
            credentials=credentials,
            discovery_timeout=discovery_timeout,
            timeout=timeout,
            on_discovered_raw=capture_raw,
        )
        click.echo(f"Detected {len(devices)} devices")
        for dev in devices.values():
            discovery_info = raw_discovery[dev.host]
            if decrypted_data := dev._discovery_info.get("decrypted_data"):
                discovery_info["result"]["decrypted_data"] = decrypted_data

            await handle_device(
                basedir,
                autosave,
                dev.protocol,
                discovery_info=discovery_info,
                batch_size=batch_size,
            )


async def get_legacy_fixture(
    protocol: IotProtocol, *, discovery_info: dict[str, dict[str, Any]] | None
) -> FixtureResult:
    """Get fixture for legacy IOT style protocol."""
    items = [
        Call(module="system", method="get_sysinfo"),
        Call(module="emeter", method="get_realtime"),
        Call(module="cnCloud", method="get_info"),
        Call(module="cnCloud", method="get_intl_fw_list"),
        Call(module="smartlife.iot.common.cloud", method="get_info"),
        Call(module="smartlife.iot.common.cloud", method="get_intl_fw_list"),
        Call(module="smartlife.iot.common.schedule", method="get_next_action"),
        Call(module="smartlife.iot.common.schedule", method="get_rules"),
        Call(module="schedule", method="get_next_action"),
        Call(module="schedule", method="get_rules"),
        Call(module="smartlife.iot.dimmer", method="get_dimmer_parameters"),
        Call(module="smartlife.iot.dimmer", method="get_default_behavior"),
        Call(module="smartlife.iot.common.emeter", method="get_realtime"),
        Call(
            module="smartlife.iot.smartbulb.lightingservice", method="get_light_state"
        ),
        Call(
            module="smartlife.iot.smartbulb.lightingservice",
            method="get_default_behavior",
        ),
        Call(
            module="smartlife.iot.smartbulb.lightingservice", method="get_light_details"
        ),
        Call(module="smartlife.iot.lightStrip", method="get_default_behavior"),
        Call(module="smartlife.iot.lightStrip", method="get_light_state"),
        Call(module="smartlife.iot.lightStrip", method="get_light_details"),
        Call(module="smartlife.iot.LAS", method="get_config"),
        Call(module="smartlife.iot.LAS", method="get_current_brt"),
        Call(module="smartlife.iot.LAS", method="get_dark_status"),
        Call(module="smartlife.iot.LAS", method="get_adc_value"),
        Call(module="smartlife.iot.PIR", method="get_config"),
        Call(module="smartlife.iot.PIR", method="get_adc_value"),
    ]

    successes = []

    for test_call in items:
        try:
            click.echo(f"Testing {test_call}..", nl=False)
            info = await protocol.query({test_call.module: {test_call.method: {}}})
            resp = info[test_call.module]
        except Exception as ex:
            click.echo(click.style(f"FAIL {ex}", fg="red"))
        else:
            if "err_msg" in resp:
                click.echo(click.style(f"FAIL {resp['err_msg']}", fg="red"))
            else:
                click.echo(click.style("OK", fg="green"))
                successes.append((test_call, info))
        finally:
            await protocol.close()

    final_query: dict = defaultdict(defaultdict)
    final: dict = defaultdict(defaultdict)
    for succ, resp in successes:
        final_query[succ.module][succ.method] = {}
        final[succ.module][succ.method] = resp

    final = default_to_regular(final)

    try:
        final = await protocol.query(final_query)
    except Exception as ex:
        _echo_error(f"Unable to query all successes at once: {ex}")
    finally:
        await protocol.close()

    final = redact_data(final, _wrap_redactors(IOT_REDACTORS))

    # Scrub the child device ids
    if children := final.get("system", {}).get("get_sysinfo", {}).get("children"):
        for index, child in enumerate(children):
            if "id" not in child:
                _LOGGER.error("Could not find a device for the child device: %s", child)
            else:
                child["id"] = f"SCRUBBED_CHILD_DEVICE_ID_{index + 1}"

    if discovery_info and not discovery_info.get("system"):
        final["discovery_result"] = redact_data(
            discovery_info, _wrap_redactors(NEW_DISCOVERY_REDACTORS)
        )

    click.echo(f"Got {len(successes)} successes")
    click.echo(click.style("## device info file ##", bold=True))

    sysinfo = final["system"]["get_sysinfo"]
    model = sysinfo["model"]
    hw_version = sysinfo["hw_ver"]
    sw_version = sysinfo["sw_ver"]
    sw_version = sw_version.split(" ", maxsplit=1)[0]
    save_filename = f"{model}_{hw_version}_{sw_version}"
    copy_folder = IOT_FOLDER
    return FixtureResult(
        filename=save_filename,
        folder=copy_folder,
        data=final,
        protocol_suffix=IOT_SUFFIX,
    )


def _echo_error(msg: str):
    click.echo(
        click.style(
            msg,
            bold=True,
            fg="red",
        )
    )


def format_exception(e):
    """Print full exception stack as if it hadn't been caught.

    https://stackoverflow.com/a/12539332
    """
    exception_list = traceback.format_stack()
    exception_list = exception_list[:-2]
    exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
    exception_list.extend(
        traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])
    )

    exception_str = "Traceback (most recent call last):\n"
    exception_str += "".join(exception_list)
    # Removing the last \n
    exception_str = exception_str[:-1]

    return exception_str


async def _make_final_calls(
    protocol: SmartProtocol,
    calls: list[SmartCall],
    name: str,
    batch_size: int,
    *,
    child_device_id: str,
) -> dict[str, dict]:
    """Call all successes again.

    After trying each call individually make the calls again either as a
    multiple request or as single requests for those that don't support
    multiple queries.
    """
    multiple_requests = {
        key: smartcall.request[key]
        for smartcall in calls
        if smartcall.supports_multiple and (key := next(iter(smartcall.request)))
    }
    final = await _make_requests_or_exit(
        protocol,
        multiple_requests,
        name + " - multiple",
        batch_size,
        child_device_id=child_device_id,
    )
    single_calls = [smartcall for smartcall in calls if not smartcall.supports_multiple]
    for smartcall in single_calls:
        final[smartcall.module] = await _make_requests_or_exit(
            protocol,
            smartcall.request,
            f"{name} + {smartcall.module}",
            batch_size,
            child_device_id=child_device_id,
        )
    return final


async def _make_requests_or_exit(
    protocol: SmartProtocol,
    requests: dict,
    name: str,
    batch_size: int,
    *,
    child_device_id: str,
) -> dict[str, dict]:
    final = {}
    # Calling close on child protocol wrappers is a noop
    protocol_to_close = protocol
    if child_device_id:
        if isinstance(protocol, SmartCamProtocol):
            protocol = _ChildCameraProtocolWrapper(child_device_id, protocol)
        else:
            protocol = _ChildProtocolWrapper(child_device_id, protocol)
    try:
        end = len(requests)
        step = batch_size  # Break the requests down as there seems to be a size limit
        keys = [key for key in requests]
        for i in range(0, end, step):
            x = i
            requests_step = {key: requests[key] for key in keys[x : x + step]}
            responses = await protocol.query(requests_step)
            for method, result in responses.items():
                final[method] = result
        return final
    except AuthenticationError as ex:
        _echo_error(
            f"Unable to query the device due to an authentication error: {ex}",
        )
        exit(1)
    except KasaException as ex:
        _echo_error(
            f"Unable to query {name} at once: {ex}",
        )
        if isinstance(ex, TimeoutError):
            _echo_error(
                "Timeout, try reducing the batch size via --batch-size option.",
            )
        exit(1)
    except Exception as ex:
        _echo_error(
            f"Unexpected exception querying {name} at once: {ex}",
        )
        if _LOGGER.isEnabledFor(logging.DEBUG):
            _echo_error(format_exception(ex))
        exit(1)
    finally:
        await protocol_to_close.close()


async def get_smart_camera_test_calls(protocol: SmartProtocol):
    """Get the list of test calls to make."""
    test_calls: list[SmartCall] = []
    successes: list[SmartCall] = []

    test_calls = []
    for request in SMARTCAM_REQUESTS:
        method = next(iter(request))
        if method == "get":
            module = method + "_" + next(iter(request[method]))
        else:
            module = method
        test_calls.append(
            SmartCall(
                module=module,
                request=request,
                should_succeed=True,
                child_device_id="",
                supports_multiple=(method != "get"),
            )
        )

    # Now get the child device requests
    child_request = {
        "getChildDeviceList": {"childControl": {"start_index": 0}},
    }
    try:
        child_response = await protocol.query(child_request)
    except Exception:
        _LOGGER.debug("Device does not have any children.")
    else:
        successes.append(
            SmartCall(
                module="getChildDeviceList",
                request=child_request,
                should_succeed=True,
                child_device_id="",
                supports_multiple=True,
            )
        )
        child_list = child_response["getChildDeviceList"]["child_device_list"]
        for child in child_list:
            child_id = child.get("device_id") or child.get("dev_id")
            if not child_id:
                _LOGGER.error("Could not find child device id in %s", child)
            # If category is in the child device map the protocol is smart.
            if (
                category := child.get("category")
            ) and category in SmartChildDevice.CHILD_DEVICE_TYPE_MAP:
                child_protocol = _ChildCameraProtocolWrapper(child_id, protocol)
                try:
                    nego_response = await child_protocol.query({"component_nego": None})
                except Exception as ex:
                    _LOGGER.error("Error calling component_nego: %s", ex)
                    continue
                if "component_nego" not in nego_response:
                    _LOGGER.error(
                        "Could not find component_nego in device response: %s",
                        nego_response,
                    )
                    continue
                successes.append(
                    SmartCall(
                        module="component_nego",
                        request={"component_nego": None},
                        should_succeed=True,
                        child_device_id=child_id,
                    )
                )
                child_components = {
                    item["id"]: item["ver_code"]
                    for item in nego_response["component_nego"]["component_list"]
                }
                for component_id, ver_code in child_components.items():
                    if (
                        requests := get_component_requests(component_id, ver_code)
                    ) is not None:
                        component_test_calls = [
                            SmartCall(
                                module=component_id,
                                request={key: val},
                                should_succeed=True,
                                child_device_id=child_id,
                            )
                            for key, val in requests.items()
                        ]
                        test_calls.extend(component_test_calls)
                    else:
                        click.echo(f"Skipping {component_id}..", nl=False)
                        click.echo(click.style("UNSUPPORTED", fg="yellow"))
            else:  # Not a smart protocol device so assume camera protocol
                for request in SMARTCAM_REQUESTS:
                    method = next(iter(request))
                    if method == "get":
                        method = method + "_" + next(iter(request[method]))
                    test_calls.append(
                        SmartCall(
                            module=method,
                            request=request,
                            should_succeed=True,
                            child_device_id=child_id,
                        )
                    )
    finally:
        await protocol.close()
    return test_calls, successes


async def get_smart_test_calls(protocol: SmartProtocol):
    """Get the list of test calls to make."""
    test_calls = []
    successes = []
    child_device_components = {}

    click.echo("Testing component_nego call ..", nl=False)
    responses = await _make_requests_or_exit(
        protocol,
        SmartRequest.component_nego().to_dict(),
        "component_nego call",
        batch_size=1,
        child_device_id="",
    )
    component_info_response = responses["component_nego"]
    click.echo(click.style("OK", fg="green"))
    successes.append(
        SmartCall(
            module="component_nego",
            request=SmartRequest("component_nego").to_dict(),
            should_succeed=True,
            child_device_id="",
        )
    )
    components = {
        item["id"]: item["ver_code"]
        for item in component_info_response["component_list"]
    }

    if "child_device" in components:
        child_components = await _make_requests_or_exit(
            protocol,
            SmartRequest.get_child_device_component_list().to_dict(),
            "child device component list",
            batch_size=1,
            child_device_id="",
        )
        successes.append(
            SmartCall(
                module="child_component_list",
                request=SmartRequest.get_child_device_component_list().to_dict(),
                should_succeed=True,
                child_device_id="",
            )
        )
        test_calls.append(
            SmartCall(
                module="child_device_list",
                request=SmartRequest.get_child_device_list().to_dict(),
                should_succeed=True,
                child_device_id="",
            )
        )
        # Get list of child components to call
        if "control_child" in components:
            child_device_components = {
                child_component_list["device_id"]: {
                    item["id"]: item["ver_code"]
                    for item in child_component_list["component_list"]
                }
                for child_component_list in child_components[
                    "get_child_device_component_list"
                ]["child_component_list"]
            }

    # Get component calls
    for component_id, ver_code in components.items():
        if component_id == "child_device":
            continue
        if (requests := get_component_requests(component_id, ver_code)) is not None:
            component_test_calls = [
                SmartCall(
                    module=component_id,
                    request={key: val},
                    should_succeed=True,
                    child_device_id="",
                )
                for key, val in requests.items()
            ]
            test_calls.extend(component_test_calls)
        else:
            click.echo(f"Skipping {component_id}..", nl=False)
            click.echo(click.style("UNSUPPORTED", fg="yellow"))

    # Child component calls
    for child_device_id, child_components in child_device_components.items():
        test_calls.append(
            SmartCall(
                module="component_nego",
                request=SmartRequest("component_nego").to_dict(),
                should_succeed=True,
                child_device_id=child_device_id,
            )
        )
        for component_id, ver_code in child_components.items():
            if (requests := get_component_requests(component_id, ver_code)) is not None:
                component_test_calls = [
                    SmartCall(
                        module=component_id,
                        request={key: val},
                        should_succeed=True,
                        child_device_id=child_device_id,
                    )
                    for key, val in requests.items()
                ]
                test_calls.extend(component_test_calls)
            else:
                click.echo(f"Skipping {component_id}..", nl=False)
                click.echo(click.style("UNSUPPORTED", fg="yellow"))

    return test_calls, successes


def get_smart_child_fixture(response, model_info, folder, suffix):
    """Get a seperate fixture for the child device."""
    hw_version = model_info.hardware_version
    fw_version = model_info.firmware_version
    model = model_info.long_name
    if model_info.region is not None:
        model = f"{model}({model_info.region})"
    save_filename = f"{model}_{hw_version}_{fw_version}"
    return FixtureResult(
        filename=save_filename,
        folder=folder,
        data=response,
        protocol_suffix=suffix,
    )


def scrub_child_device_ids(
    main_response: dict, child_responses: dict
) -> dict[str, str]:
    """Scrub all the child device ids in the responses."""
    # Make the scrubbed id map
    scrubbed_child_id_map = {
        device_id: f"SCRUBBED_CHILD_DEVICE_ID_{index + 1}"
        for index, device_id in enumerate(child_responses.keys())
        if device_id != ""
    }

    for child_id, response in child_responses.items():
        scrubbed_child_id = scrubbed_child_id_map[child_id]
        # scrub the device id in the child's get info response
        # The checks for the device_id will ensure we can get a fixture
        # even if the data is unexpectedly not available although it should
        # always be there
        if "get_device_info" in response and "device_id" in response["get_device_info"]:
            response["get_device_info"]["device_id"] = scrubbed_child_id
        elif (
            basic_info := response.get("getDeviceInfo", {})
            .get("device_info", {})
            .get("basic_info")
        ) and "dev_id" in basic_info:
            basic_info["dev_id"] = scrubbed_child_id
        else:
            _LOGGER.error(
                "Cannot find device id in child get device info: %s", child_id
            )

    # Scrub the device ids in the parent for smart protocol
    if gc := main_response.get("get_child_device_component_list"):
        for child in gc["child_component_list"]:
            device_id = child["device_id"]
            child["device_id"] = scrubbed_child_id_map[device_id]
        for child in main_response["get_child_device_list"]["child_device_list"]:
            device_id = child["device_id"]
            child["device_id"] = scrubbed_child_id_map[device_id]

    # Scrub the device ids in the parent for the smart camera protocol
    if gc := main_response.get("getChildDeviceComponentList"):
        for child in gc["child_component_list"]:
            device_id = child["device_id"]
            child["device_id"] = scrubbed_child_id_map[device_id]
        for child in main_response["getChildDeviceList"]["child_device_list"]:
            if device_id := child.get("device_id"):
                child["device_id"] = scrubbed_child_id_map[device_id]
                continue
            elif dev_id := child.get("dev_id"):
                child["dev_id"] = scrubbed_child_id_map[dev_id]
                continue
            _LOGGER.error("Could not find a device id for the child device: %s", child)

    return scrubbed_child_id_map


async def get_smart_fixtures(
    protocol: SmartProtocol,
    *,
    discovery_info: dict[str, dict[str, Any]] | None,
    batch_size: int,
) -> list[FixtureResult]:
    """Get fixture for new TAPO style protocol."""
    if isinstance(protocol, SmartCamProtocol):
        test_calls, successes = await get_smart_camera_test_calls(protocol)
        child_wrapper: type[_ChildProtocolWrapper | _ChildCameraProtocolWrapper] = (
            _ChildCameraProtocolWrapper
        )
    else:
        test_calls, successes = await get_smart_test_calls(protocol)
        child_wrapper = _ChildProtocolWrapper

    for test_call in test_calls:
        click.echo(f"Testing  {test_call.module}..", nl=False)
        try:
            click.echo(f"Testing {test_call}..", nl=False)
            if test_call.child_device_id == "":
                response = await protocol.query(test_call.request)
            else:
                cp = child_wrapper(test_call.child_device_id, protocol)
                response = await cp.query(test_call.request)
        except AuthenticationError as ex:
            _echo_error(
                f"Unable to query the device due to an authentication error: {ex}",
            )
            exit(1)
        except Exception as ex:
            if (
                not test_call.should_succeed
                and hasattr(ex, "error_code")
                and ex.error_code
                in [
                    SmartErrorCode.UNKNOWN_METHOD_ERROR,
                    SmartErrorCode.TRANSPORT_NOT_AVAILABLE_ERROR,
                    SmartErrorCode.UNSPECIFIC_ERROR,
                ]
            ):
                click.echo(click.style("FAIL - EXPECTED", fg="green"))
            else:
                click.echo(click.style(f"FAIL {ex}", fg="red"))
        else:
            if not response:
                click.echo(click.style("FAIL no response", fg="red"))
            else:
                if not test_call.should_succeed:
                    click.echo(click.style("OK - EXPECTED FAIL", fg="red"))
                else:
                    click.echo(click.style("OK", fg="green"))
                successes.append(test_call)
        finally:
            await protocol.close()

    # Put all the successes into a dict[child_device_id or "", successes[]]
    device_requests: dict[str, list[SmartCall]] = {}
    for success in successes:
        device_request = device_requests.setdefault(success.child_device_id, [])
        device_request.append(success)

    final = await _make_final_calls(
        protocol, device_requests[""], "All successes", batch_size, child_device_id=""
    )
    fixture_results = []

    # Make the final child calls
    child_responses = {}
    for child_device_id, requests in device_requests.items():
        if child_device_id == "":
            continue
        response = await _make_final_calls(
            protocol,
            requests,
            "All child successes",
            batch_size,
            child_device_id=child_device_id,
        )
        child_responses[child_device_id] = response

    # scrub the child ids
    scrubbed_child_id_map = scrub_child_device_ids(final, child_responses)

    # Redact data from the main device response. _wrap_redactors ensure we do
    # not redact the scrubbed child device ids and replaces REDACTED_partial_id
    # with zeros
    final = redact_data(final, _wrap_redactors(SMART_REDACTORS))

    # smart cam child devices provide more information in getChildDeviceList on the
    # parent than they return when queried directly for getDeviceInfo so we will store
    # it in the child fixture.
    if smart_cam_child_list := final.get("getChildDeviceList"):
        child_infos_on_parent = {
            info["device_id"]: info
            for info in smart_cam_child_list["child_device_list"]
        }

    for child_id, response in child_responses.items():
        scrubbed_child_id = scrubbed_child_id_map[child_id]

        # Get the parent model for checking whether to create a seperate child fixture
        if model := final.get("get_device_info", {}).get("model"):
            parent_model = model
        elif (
            device_model := final.get("getDeviceInfo", {})
            .get("device_info", {})
            .get("basic_info", {})
            .get("device_model")
        ):
            parent_model = device_model
        else:
            parent_model = None
            _LOGGER.error("Cannot determine parent device model.")

        # different model smart child device
        if (
            (child_model := response.get("get_device_info", {}).get("model"))
            and parent_model
            and child_model != parent_model
        ):
            response = redact_data(response, _wrap_redactors(SMART_REDACTORS))
            model_info = SmartDevice._get_device_info(response, None)
            fixture_results.append(
                get_smart_child_fixture(
                    response, model_info, SMART_CHILD_FOLDER, SMART_CHILD_SUFFIX
                )
            )
        # different model smartcam child device
        elif (
            (
                child_model := response.get("getDeviceInfo", {})
                .get("device_info", {})
                .get("basic_info", {})
                .get("device_model")
            )
            and parent_model
            and child_model != parent_model
        ):
            response = redact_data(response, _wrap_redactors(SMART_REDACTORS))
            # There is more info in the childDeviceList on the parent
            # particularly the region is needed here.
            child_info_from_parent = child_infos_on_parent[scrubbed_child_id]
            response[CHILD_INFO_FROM_PARENT] = child_info_from_parent
            model_info = SmartCamChild._get_device_info(response, None)
            fixture_results.append(
                get_smart_child_fixture(
                    response, model_info, SMARTCAM_CHILD_FOLDER, SMARTCAM_CHILD_SUFFIX
                )
            )
        # same model child device
        else:
            cd = final.setdefault("child_devices", {})
            cd[scrubbed_child_id] = response

    discovery_result = None
    if discovery_info:
        final["discovery_result"] = redact_data(
            discovery_info, _wrap_redactors(NEW_DISCOVERY_REDACTORS)
        )
        discovery_result = discovery_info["result"]

    click.echo(f"Got {len(successes)} successes")
    click.echo(click.style("## device info file ##", bold=True))

    if "get_device_info" in final:
        # smart protocol
        model_info = SmartDevice._get_device_info(final, discovery_result)
        copy_folder = SMART_FOLDER
        protocol_suffix = SMART_PROTOCOL_SUFFIX
    else:
        # smart camera protocol
        model_info = SmartCamDevice._get_device_info(final, discovery_result)
        copy_folder = SMARTCAM_FOLDER
        protocol_suffix = SMARTCAM_SUFFIX
    hw_version = model_info.hardware_version
    sw_version = model_info.firmware_version
    model = model_info.long_name
    if model_info.region is not None:
        model = f"{model}({model_info.region})"

    save_filename = f"{model}_{hw_version}_{sw_version}"

    fixture_results.insert(
        0,
        FixtureResult(
            filename=save_filename,
            folder=copy_folder,
            data=final,
            protocol_suffix=protocol_suffix,
        ),
    )
    return fixture_results


if __name__ == "__main__":
    cli()



"""Implementation of cloud module."""

from __future__ import annotations

from ...feature import Feature
from ..smartmodule import SmartModule


class Cloud(SmartModule):
    """Implementation of cloud module."""

    QUERY_GETTER_NAME = "get_connect_cloud_state"
    REQUIRED_COMPONENT = "cloud_connect"
    MINIMUM_UPDATE_INTERVAL_SECS = 60

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="cloud_connection",
                name="Cloud connection",
                container=self,
                attribute_getter="is_connected",
                icon="mdi:cloud",
                type=Feature.Type.BinarySensor,
                category=Feature.Category.Info,
            )
        )

    @property
    def is_connected(self) -> bool:
        """Return True if device is connected to the cloud."""
        if self._has_data_error():
            return False
        return self.data["status"] == 0



import copy
import logging
from contextlib import nullcontext as does_not_raise
from unittest.mock import patch

import pytest

from kasa import DeviceError, Module
from kasa.exceptions import SmartErrorCode
from kasa.interfaces.energy import Energy
from kasa.smart import SmartDevice
from kasa.smart.modules import Energy as SmartEnergyModule
from tests.conftest import has_emeter_smart


@has_emeter_smart
async def test_supported(dev: SmartDevice):
    energy_module = dev.modules.get(Module.Energy)
    if not energy_module:
        pytest.skip(f"Energy module not supported for {dev}.")

    assert isinstance(energy_module, SmartEnergyModule)
    assert energy_module.supports(Energy.ModuleFeature.CONSUMPTION_TOTAL) is False
    assert energy_module.supports(Energy.ModuleFeature.PERIODIC_STATS) is False
    if energy_module.supported_version < 2:
        assert energy_module.supports(Energy.ModuleFeature.VOLTAGE_CURRENT) is False
    else:
        assert energy_module.supports(Energy.ModuleFeature.VOLTAGE_CURRENT) is True


@has_emeter_smart
async def test_get_energy_usage_error(
    dev: SmartDevice, caplog: pytest.LogCaptureFixture
):
    """Test errors on get_energy_usage."""
    caplog.set_level(logging.DEBUG)

    energy_module = dev.modules.get(Module.Energy)
    if not energy_module:
        pytest.skip(f"Energy module not supported for {dev}.")

    version = dev._components["energy_monitoring"]

    expected_raise = does_not_raise() if version > 1 else pytest.raises(DeviceError)
    if version > 1:
        expected = "get_energy_usage"
        expected_current_consumption = 2.002
    else:
        expected = "current_power"
        expected_current_consumption = None

    assert expected in energy_module.data
    assert energy_module.current_consumption is not None
    assert energy_module.consumption_today is not None
    assert energy_module.consumption_this_month is not None

    last_update = copy.deepcopy(dev._last_update)
    resp = copy.deepcopy(last_update)

    if ed := resp.get("get_emeter_data"):
        ed["power_mw"] = 2002
    if cp := resp.get("get_current_power"):
        cp["current_power"] = 2.002
    resp["get_energy_usage"] = SmartErrorCode.JSON_DECODE_FAIL_ERROR

    # version 1 only has get_energy_usage so module should raise an error if
    # version 1 and get_energy_usage is in error
    with patch.object(dev.protocol, "query", return_value=resp):
        await dev.update()

    with expected_raise:
        assert "get_energy_usage" not in energy_module.data

    assert energy_module.current_consumption == expected_current_consumption
    assert energy_module.consumption_today is None
    assert energy_module.consumption_this_month is None

    msg = (
        f"Removed key get_energy_usage from response for device {dev.host}"
        " as it returned error: JSON_DECODE_FAIL_ERROR"
    )
    if version > 1:
        assert msg in caplog.text

    # Now test with no get_emeter_data
    # This may not be valid scenario but we have a fallback to get_current_power
    # just in case that should be tested.
    caplog.clear()
    resp = copy.deepcopy(last_update)

    if cp := resp.get("get_current_power"):
        cp["current_power"] = 2.002
    resp["get_energy_usage"] = SmartErrorCode.JSON_DECODE_FAIL_ERROR

    # Remove get_emeter_data from the response and from the device which will
    # remember it otherwise.
    resp.pop("get_emeter_data", None)
    dev._last_update.pop("get_emeter_data", None)

    with patch.object(dev.protocol, "query", return_value=resp):
        await dev.update()

    with expected_raise:
        assert "get_energy_usage" not in energy_module.data

    assert energy_module.current_consumption == expected_current_consumption

    # message should only be logged once
    assert msg not in caplog.text



"""Implementation of brightness module."""

from __future__ import annotations

from ...feature import Feature
from ..smartmodule import Module, SmartModule

BRIGHTNESS_MIN = 0
BRIGHTNESS_MAX = 100


class Brightness(SmartModule):
    """Implementation of brightness module."""

    REQUIRED_COMPONENT = "brightness"

    def _initialize_features(self) -> None:
        """Initialize features."""
        super()._initialize_features()

        device = self._device
        self._add_feature(
            Feature(
                device,
                id="brightness",
                name="Brightness",
                container=self,
                attribute_getter="brightness",
                attribute_setter="set_brightness",
                range_getter=lambda: (BRIGHTNESS_MIN, BRIGHTNESS_MAX),
                type=Feature.Type.Number,
                category=Feature.Category.Primary,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        # Brightness is contained in the main device info response.
        return {}

    @property
    def brightness(self) -> int:
        """Return current brightness."""
        # If the device supports effects and one is active, use its brightness
        if (
            light_effect := self._device.modules.get(Module.SmartLightEffect)
        ) is not None and light_effect.is_active:
            return light_effect.brightness

        return self.data["brightness"]

    async def set_brightness(
        self, brightness: int, *, transition: int | None = None
    ) -> dict:
        """Set the brightness. A brightness value of 0 will turn off the light.

        Note, transition is not supported and will be ignored.
        """
        if not isinstance(brightness, int) or not (
            BRIGHTNESS_MIN <= brightness <= BRIGHTNESS_MAX
        ):
            raise ValueError(
                f"Invalid brightness value: {brightness} "
                f"(valid range: {BRIGHTNESS_MIN}-{BRIGHTNESS_MAX}%)"
            )

        if brightness == 0:
            return await self._device.turn_off()

        # If the device supports effects and one is active, we adjust its brightness
        if (
            light_effect := self._device.modules.get(Module.SmartLightEffect)
        ) is not None and light_effect.is_active:
            return await light_effect.set_brightness(brightness)

        return await self.call("set_device_info", {"brightness": brightness})

    async def _check_supported(self) -> bool:
        """Additional check to see if the module is supported by the device."""
        return "brightness" in self.data



"""Power protection module."""

from __future__ import annotations

from typing import Annotated

from ...feature import Feature
from ...module import FeatureAttribute
from ..smartmodule import SmartModule


class PowerProtection(SmartModule):
    """Implementation for power_protection."""

    REQUIRED_COMPONENT = "power_protection"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                device=self._device,
                id="overloaded",
                name="Overloaded",
                container=self,
                attribute_getter="overloaded",
                type=Feature.Type.BinarySensor,
                category=Feature.Category.Info,
            )
        )
        self._add_feature(
            Feature(
                device=self._device,
                id="power_protection_threshold",
                name="Power protection threshold",
                container=self,
                attribute_getter="_threshold_or_zero",
                attribute_setter="_set_threshold_auto_enable",
                unit_getter=lambda: "W",
                type=Feature.Type.Number,
                range_getter=lambda: (0, self._max_power),
                category=Feature.Category.Config,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {"get_protection_power": {}, "get_max_power": {}}

    @property
    def overloaded(self) -> bool:
        """Return True is power protection has been triggered.

        This value remains True until the device is turned on again.
        """
        return self._device.sys_info["power_protection_status"] == "overloaded"

    @property
    def enabled(self) -> bool:
        """Return True if child protection is enabled."""
        return self.data["get_protection_power"]["enabled"]

    async def set_enabled(self, enabled: bool, *, threshold: int | None = None) -> dict:
        """Set power protection enabled.

        If power protection has never been enabled before the threshold will
        be 0 so if threshold is not provided it will be set to half the max.
        """
        if threshold is None and enabled and self.protection_threshold == 0:
            threshold = int(self._max_power / 2)

        if threshold and (threshold < 0 or threshold > self._max_power):
            raise ValueError(
                "Threshold out of range: %s (%s)", threshold, self.protection_threshold
            )

        params = {**self.data["get_protection_power"], "enabled": enabled}
        if threshold is not None:
            params["protection_power"] = threshold
        return await self.call("set_protection_power", params)

    async def _set_threshold_auto_enable(self, threshold: int) -> dict:
        """Set power protection and enable."""
        if threshold == 0:
            return await self.set_enabled(False)
        else:
            return await self.set_enabled(True, threshold=threshold)

    @property
    def _threshold_or_zero(self) -> int:
        """Get power protection threshold. 0 if not enabled."""
        return self.protection_threshold if self.enabled else 0

    @property
    def _max_power(self) -> int:
        """Return max power."""
        return self.data["get_max_power"]["max_power"]

    @property
    def protection_threshold(
        self,
    ) -> Annotated[int, FeatureAttribute("power_protection_threshold")]:
        """Return protection threshold in watts."""
        # If never configured, there is no value set.
        return self.data["get_protection_power"].get("protection_power", 0)

    async def set_protection_threshold(self, threshold: int) -> dict:
        """Set protection threshold."""
        if threshold < 0 or threshold > self._max_power:
            raise ValueError(
                "Threshold out of range: %s (%s)", threshold, self.protection_threshold
            )

        params = {
            **self.data["get_protection_power"],
            "protection_power": threshold,
        }
        return await self.call("set_protection_power", params)

    async def _check_supported(self) -> bool:
        """Return True if module is supported.

        This is needed, as strips like P304M report the status only for children.
        """
        return "power_protection_status" in self._device.sys_info



"""Module for cli usage commands.."""

from __future__ import annotations

from typing import cast

import asyncclick as click

from kasa import (
    Device,
    Module,
)
from kasa.interfaces import Energy
from kasa.iot.modules import Usage

from .common import (
    echo,
    error,
    pass_dev_or_child,
)


@click.command()
@click.option("--year", type=click.DateTime(["%Y"]), default=None, required=False)
@click.option("--month", type=click.DateTime(["%Y-%m"]), default=None, required=False)
@click.option("--erase", is_flag=True)
@pass_dev_or_child
async def energy(dev: Device, year, month, erase):
    """Query energy module for historical consumption.

    Daily and monthly data provided in CSV format.
    """
    echo("[bold]== Energy ==[/bold]")
    if not (energy := dev.modules.get(Module.Energy)):
        error("Device has no energy module.")
        return

    if (year or month or erase) and not energy.supports(
        Energy.ModuleFeature.PERIODIC_STATS
    ):
        error("Device does not support historical statistics")
        return

    if erase:
        echo("Erasing emeter statistics..")
        return await energy.erase_stats()

    if year:
        echo(f"== For year {year.year} ==")
        echo("Month, usage (kWh)")
        usage_data = await energy.get_monthly_stats(year=year.year)
    elif month:
        echo(f"== For month {month.month} of {month.year} ==")
        echo("Day, usage (kWh)")
        usage_data = await energy.get_daily_stats(year=month.year, month=month.month)
    else:
        # Call with no argument outputs summary data and returns
        emeter_status = energy.status

        echo("Current: {} A".format(emeter_status["current"]))
        echo("Voltage: {} V".format(emeter_status["voltage"]))
        echo("Power: {} W".format(emeter_status["power"]))
        echo("Total consumption: {} kWh".format(emeter_status["total"]))

        echo(f"Today: {energy.consumption_today} kWh")
        echo(f"This month: {energy.consumption_this_month} kWh")

        return emeter_status

    # output any detailed usage data
    for index, usage in usage_data.items():
        echo(f"{index}, {usage}")

    return usage_data


@click.command()
@click.option("--year", type=click.DateTime(["%Y"]), default=None, required=False)
@click.option("--month", type=click.DateTime(["%Y-%m"]), default=None, required=False)
@click.option("--erase", is_flag=True)
@pass_dev_or_child
async def usage(dev: Device, year, month, erase):
    """Query usage for historical consumption.

    Daily and monthly data provided in CSV format.
    """
    echo("[bold]== Usage ==[/bold]")
    usage = cast(Usage, dev.modules["usage"])

    if erase:
        echo("Erasing usage statistics..")
        return await usage.erase_stats()

    if year:
        echo(f"== For year {year.year} ==")
        echo("Month, usage (minutes)")
        usage_data = await usage.get_monthstat(year=year.year)
    elif month:
        echo(f"== For month {month.month} of {month.year} ==")
        echo("Day, usage (minutes)")
        usage_data = await usage.get_daystat(year=month.year, month=month.month)
    else:
        # Call with no argument outputs summary data and returns
        echo(f"Today: {usage.usage_today} minutes")
        echo(f"This month: {usage.usage_this_month} minutes")

        return usage

    # output any detailed usage data
    for index, usage in usage_data.items():
        echo(f"{index}, {usage}")

    return usage_data



"""Module for emeter container."""

from __future__ import annotations

import logging

_LOGGER = logging.getLogger(__name__)


class EmeterStatus(dict):
    """Container for converting different representations of emeter data.

    Newer FW/HW versions postfix the variable names with the used units,
    where-as the olders do not have this feature.

    This class automatically converts between these two to allow
    backwards and forwards compatibility.
    """

    @property
    def voltage(self) -> float | None:
        """Return voltage in V."""
        try:
            return self["voltage"]
        except ValueError:
            return None

    @property
    def power(self) -> float | None:
        """Return power in W."""
        try:
            return self["power"]
        except ValueError:
            return None

    @property
    def current(self) -> float | None:
        """Return current in A."""
        try:
            return self["current"]
        except ValueError:
            return None

    @property
    def total(self) -> float | None:
        """Return total in kWh."""
        try:
            return self["total"]
        except ValueError:
            return None

    def __repr__(self) -> str:
        return (
            f"<EmeterStatus power={self.power} voltage={self.voltage}"
            f" current={self.current} total={self.total}>"
        )

    def __getitem__(self, item: str) -> float | None:
        """Return value in wanted units."""
        valid_keys = [
            "voltage_mv",
            "power_mw",
            "current_ma",
            "energy_wh",
            "total_wh",
            "voltage",
            "power",
            "current",
            "total",
            "energy",
        ]

        # 1. if requested data is available, return it
        if item in super().keys():  # noqa: SIM118
            return super().__getitem__(item)
        # otherwise decide how to convert it
        else:
            if item not in valid_keys:
                raise KeyError(item)
            if "_" in item:  # upscale
                return super().__getitem__(item[: item.find("_")]) * 1000
            else:  # downscale
                for i in super().keys():  # noqa: SIM118
                    if (
                        i.startswith(item)
                        and (value := self.__getitem__(i)) is not None
                    ):
                        return value / 1000

                _LOGGER.debug("Unable to find value for '%s'", item)
                return None



"""Module for cli vacuum commands.."""

from __future__ import annotations

import asyncclick as click

from kasa import (
    Device,
    Module,
)

from .common import (
    error,
    pass_dev_or_child,
)


@click.group(invoke_without_command=False)
@click.pass_context
async def vacuum(ctx: click.Context) -> None:
    """Vacuum commands."""


@vacuum.group(invoke_without_command=True, name="records")
@pass_dev_or_child
async def records_group(dev: Device) -> None:
    """Access cleaning records."""
    if not (rec := dev.modules.get(Module.CleanRecords)):
        error("This device does not support records.")

    data = rec.parsed_data
    latest = data.last_clean
    click.echo(
        f"Totals: {rec.total_clean_area} {rec.area_unit} in {rec.total_clean_time} "
        f"(cleaned {rec.total_clean_count} times)"
    )
    click.echo(f"Last clean: {latest.clean_area} {rec.area_unit} @ {latest.clean_time}")
    click.echo("Execute `kasa vacuum records list` to list all records.")


@records_group.command(name="list")
@pass_dev_or_child
async def records_list(dev: Device) -> None:
    """List all cleaning records."""
    if not (rec := dev.modules.get(Module.CleanRecords)):
        error("This device does not support records.")

    data = rec.parsed_data
    for record in data.records:
        click.echo(
            f"* {record.timestamp}: cleaned {record.clean_area} {rec.area_unit}"
            f" in {record.clean_time}"
        )


@vacuum.group(invoke_without_command=True, name="consumables")
@pass_dev_or_child
@click.pass_context
async def consumables(ctx: click.Context, dev: Device) -> None:
    """List device consumables."""
    if not (cons := dev.modules.get(Module.Consumables)):
        error("This device does not support consumables.")

    if not ctx.invoked_subcommand:
        for c in cons.consumables.values():
            click.echo(f"{c.name} ({c.id}): {c.used} used, {c.remaining} remaining")


@consumables.command(name="reset")
@click.argument("consumable_id", required=True)
@pass_dev_or_child
async def reset_consumable(dev: Device, consumable_id: str) -> None:
    """Reset the consumable used/remaining time."""
    cons = dev.modules[Module.Consumables]

    if consumable_id not in cons.consumables:
        error(
            f"Consumable {consumable_id} not found in "
            f"device consumables: {', '.join(cons.consumables.keys())}."
        )

    await cons.reset_consumable(consumable_id)

    click.echo(f"Consumable {consumable_id} reset")



"""Implementation of auto off module."""

from __future__ import annotations

import logging
from datetime import datetime, timedelta

from ...feature import Feature
from ..smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class AutoOff(SmartModule):
    """Implementation of auto off module."""

    REQUIRED_COMPONENT = "auto_off"
    QUERY_GETTER_NAME = "get_auto_off_config"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="auto_off_enabled",
                name="Auto off enabled",
                container=self,
                attribute_getter="enabled",
                attribute_setter="set_enabled",
                type=Feature.Type.Switch,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="auto_off_minutes",
                name="Auto off in",
                container=self,
                attribute_getter="delay",
                attribute_setter="set_delay",
                type=Feature.Type.Number,
                unit_getter=lambda: "min",  # ha-friendly unit, see UnitOfTime.MINUTES
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="auto_off_at",
                name="Auto off at",
                container=self,
                attribute_getter="auto_off_at",
                category=Feature.Category.Info,
                type=Feature.Type.Sensor,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {self.QUERY_GETTER_NAME: {"start_index": 0}}

    @property
    def enabled(self) -> bool:
        """Return True if enabled."""
        return self.data["enable"]

    async def set_enabled(self, enable: bool) -> dict:
        """Enable/disable auto off."""
        return await self.call(
            "set_auto_off_config",
            {"enable": enable, "delay_min": self.data["delay_min"]},
        )

    @property
    def delay(self) -> int:
        """Return time until auto off."""
        return self.data["delay_min"]

    async def set_delay(self, delay: int) -> dict:
        """Set time until auto off."""
        return await self.call(
            "set_auto_off_config", {"delay_min": delay, "enable": self.data["enable"]}
        )

    @property
    def is_timer_active(self) -> bool:
        """Return True is auto-off timer is active."""
        return self._device.sys_info["auto_off_status"] == "on"

    @property
    def auto_off_at(self) -> datetime | None:
        """Return when the device will be turned off automatically."""
        if not self.is_timer_active:
            return None

        sysinfo = self._device.sys_info

        return self._device.time + timedelta(seconds=sysinfo["auto_off_remain_time"])

    async def _check_supported(self) -> bool:
        """Additional check to see if the module is supported by the device.

        Parent devices that report components of children such as P300 will not have
        the auto_off_status is sysinfo.
        """
        return "auto_off_status" in self._device.sys_info



"""Module for lazily instantiating sub modules.

Taken from the click help files.
"""

from __future__ import annotations

import importlib

import asyncclick as click


class LazyGroup(click.Group):
    """Lazy group class."""

    def __init__(self, *args, lazy_subcommands=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # lazy_subcommands is a map of the form:
        #
        #   {command-name} -> {module-name}.{command-object-name}
        #
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx):
        """List click commands."""
        base = super().list_commands(ctx)
        lazy = list(self.lazy_subcommands.keys())
        return lazy + base

    def get_command(self, ctx, cmd_name):
        """Get click command."""
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def format_commands(self, ctx, formatter) -> None:
        """Format the top level help output."""
        sections: dict[str, list] = {}
        for cmd, parent in self.lazy_subcommands.items():
            sections.setdefault(parent, [])
            cmd_obj = self.get_command(ctx, cmd)
            help = cmd_obj.get_short_help_str()
            sections[parent].append((cmd, help))
        for section in sections:
            if section:
                header = (
                    f"Common {section} commands (also available "
                    f"under the `{section}` subcommand)"
                )
            else:
                header = "Subcommands"
            with formatter.section(header):
                formatter.write_dl(sections[section])

    def _lazy_load(self, cmd_name):
        # lazily loading a command, first get the module name and attribute name
        if not (import_path := self.lazy_subcommands[cmd_name]):
            import_path = f".{cmd_name}.{cmd_name}"
        else:
            import_path = f".{import_path}.{cmd_name}"
        modname, cmd_object_name = import_path.rsplit(".", 1)
        # do the import
        mod = importlib.import_module(modname, package=__package__)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd_object_name)
        # check the result to make debugging easier
        if not isinstance(cmd_object, click.BaseCommand):
            raise ValueError(
                f"Lazy loading of {cmd_name} failed by returning a non-command object"
            )
        return cmd_object



import asyncio

import pytest
import xdoctest

from .conftest import (
    get_device_for_fixture_protocol,
    get_fixture_info,
    patch_discovery,
)


def test_bulb_examples(mocker):
    """Use KL130 (bulb with all features) to test the doctests."""
    p = asyncio.run(get_device_for_fixture_protocol("KL130(US)_1.0_1.8.11.json", "IOT"))
    mocker.patch("kasa.iot.iotbulb.IotBulb", return_value=p)
    mocker.patch("kasa.iot.iotbulb.IotBulb.update")
    res = xdoctest.doctest_module("kasa.iot.iotbulb", "all")
    assert not res["failed"]


def test_iotdevice_examples(mocker):
    """Use HS110 for emeter examples."""
    p = asyncio.run(get_device_for_fixture_protocol("HS110(EU)_1.0_1.2.5.json", "IOT"))
    asyncio.run(p.set_alias("Bedroom Lamp Plug"))
    asyncio.run(p.update())

    mocker.patch("kasa.iot.iotdevice.IotDevice", return_value=p)
    mocker.patch("kasa.iot.iotdevice.IotDevice.update")
    res = xdoctest.doctest_module("kasa.iot.iotdevice", "all")
    assert not res["failed"]


def test_plug_examples(mocker):
    """Test plug examples."""
    p = asyncio.run(get_device_for_fixture_protocol("HS110(EU)_1.0_1.2.5.json", "IOT"))
    asyncio.run(p.set_alias("Bedroom Lamp Plug"))
    asyncio.run(p.update())
    mocker.patch("kasa.iot.iotplug.IotPlug", return_value=p)
    mocker.patch("kasa.iot.iotplug.IotPlug.update")
    res = xdoctest.doctest_module("kasa.iot.iotplug", "all")
    assert not res["failed"]


def test_strip_examples(readmes_mock):
    """Test strip examples."""
    res = xdoctest.doctest_module("kasa.iot.iotstrip", "all")
    assert not res["failed"]


def test_dimmer_examples(mocker):
    """Test dimmer examples."""
    p = asyncio.run(get_device_for_fixture_protocol("HS220(US)_1.0_1.5.7.json", "IOT"))
    mocker.patch("kasa.iot.iotdimmer.IotDimmer", return_value=p)
    mocker.patch("kasa.iot.iotdimmer.IotDimmer.update")
    res = xdoctest.doctest_module("kasa.iot.iotdimmer", "all")
    assert not res["failed"]


def test_lightstrip_examples(mocker):
    """Test lightstrip examples."""
    p = asyncio.run(get_device_for_fixture_protocol("KL430(US)_1.0_1.0.10.json", "IOT"))
    asyncio.run(p.set_alias("Bedroom Lightstrip"))
    asyncio.run(p.update())
    mocker.patch("kasa.iot.iotlightstrip.IotLightStrip", return_value=p)
    mocker.patch("kasa.iot.iotlightstrip.IotLightStrip.update")
    res = xdoctest.doctest_module("kasa.iot.iotlightstrip", "all")
    assert not res["failed"]


def test_discovery_examples(readmes_mock):
    """Test discovery examples."""
    res = xdoctest.doctest_module("kasa.discover", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_deviceconfig_examples(readmes_mock):
    """Test discovery examples."""
    res = xdoctest.doctest_module("kasa.deviceconfig", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_device_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.device", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_light_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.interfaces.light", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_light_preset_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.interfaces.lightpreset", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_light_effect_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.interfaces.lighteffect", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_child_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.smart.modules.childdevice", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_module_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.module", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_feature_examples(readmes_mock):
    """Test device examples."""
    res = xdoctest.doctest_module("kasa.feature", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_tutorial_examples(readmes_mock):
    """Test discovery examples."""
    res = xdoctest.doctest_module("docs/tutorial.py", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


def test_childsetup_examples(readmes_mock, mocker):
    """Test device examples."""
    pair_resp = [
        {
            "device_id": "SCRUBBED_CHILD_DEVICE_ID_5",
            "category": "subg.trigger.button",
            "device_model": "S200B",
            "name": "I01BU0tFRF9OQU1FIw====",
        }
    ]
    mocker.patch(
        "kasa.smartcam.modules.childsetup.ChildSetup.pair", return_value=pair_resp
    )
    res = xdoctest.doctest_module("kasa.interfaces.childsetup", "all")
    assert res["n_passed"] > 0
    assert res["n_warned"] == 0
    assert not res["failed"]


@pytest.fixture
async def readmes_mock(mocker):
    fixture_infos = {
        "127.0.0.1": get_fixture_info("KP303(UK)_1.0_1.0.3.json", "IOT"),  # Strip
        "127.0.0.2": get_fixture_info("HS110(EU)_1.0_1.2.5.json", "IOT"),  # Plug
        "127.0.0.3": get_fixture_info("L530E(EU)_3.0_1.1.6.json", "SMART"),  # Bulb
        "127.0.0.4": get_fixture_info("KL430(US)_1.0_1.0.10.json", "IOT"),  # Lightstrip
        "127.0.0.5": get_fixture_info("HS220(US)_1.0_1.5.7.json", "IOT"),  # Dimmer
        "127.0.0.6": get_fixture_info("H200(US)_1.0_1.3.6.json", "SMARTCAM"),  # Hub
    }
    fixture_infos["127.0.0.1"].data["system"]["get_sysinfo"]["alias"] = (
        "Bedroom Power Strip"
    )
    for index, child in enumerate(
        fixture_infos["127.0.0.1"].data["system"]["get_sysinfo"]["children"]
    ):
        child["alias"] = f"Plug {index + 1}"
    fixture_infos["127.0.0.2"].data["system"]["get_sysinfo"]["alias"] = (
        "Bedroom Lamp Plug"
    )
    fixture_infos["127.0.0.3"].data["get_device_info"]["nickname"] = (
        "TGl2aW5nIFJvb20gQnVsYg=="  # Living Room Bulb
    )
    fixture_infos["127.0.0.4"].data["system"]["get_sysinfo"]["alias"] = (
        "Bedroom Lightstrip"
    )
    fixture_infos["127.0.0.5"].data["system"]["get_sysinfo"]["alias"] = (
        "Living Room Dimmer Switch"
    )
    fixture_infos["127.0.0.6"].data["getDeviceInfo"]["device_info"]["basic_info"][
        "device_alias"
    ] = "Tapo Hub"
    return patch_discovery(fixture_infos, mocker)



"""Interact with a TPLink Light.

>>> from kasa import Discover, Module
>>>
>>> dev = await Discover.discover_single(
>>>     "127.0.0.3",
>>>     username="user@example.com",
>>>     password="great_password"
>>> )
>>> await dev.update()
>>> print(dev.alias)
Living Room Bulb

Lights, like any other supported devices, can be turned on and off:

>>> print(dev.is_on)
>>> await dev.turn_on()
>>> await dev.update()
>>> print(dev.is_on)
True

Get the light module to interact:

>>> light = dev.modules[Module.Light]

You can use the ``has_feature()`` method to check for supported features:

>>> light.has_feature("brightness")
True
>>> light.has_feature("hsv")
True
>>> light.has_feature("color_temp")
True

All known bulbs support changing the brightness:

>>> light.brightness
100
>>> await light.set_brightness(50)
>>> await dev.update()
>>> light.brightness
50

Bulbs supporting color temperature can be queried for the supported range:

>>> if color_temp_feature := light.get_feature("color_temp"):
>>>     print(f"{color_temp_feature.minimum_value}, {color_temp_feature.maximum_value}")
2500, 6500
>>> await light.set_color_temp(3000)
>>> await dev.update()
>>> light.color_temp
3000

Color bulbs can be adjusted by passing hue, saturation and value:

>>> await light.set_hsv(180, 100, 80)
>>> await dev.update()
>>> light.hsv
HSV(hue=180, saturation=100, value=80)


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Any, NamedTuple
from warnings import warn

from ..exceptions import KasaException
from ..module import FeatureAttribute, Module


@dataclass
class LightState:
    """Class for smart light preset info."""

    light_on: bool | None = None
    brightness: int | None = None
    hue: int | None = None
    saturation: int | None = None
    color_temp: int | None = None
    transition: int | None = None


class ColorTempRange(NamedTuple):
    """Color temperature range."""

    min: int
    max: int


class HSV(NamedTuple):
    """Hue-saturation-value."""

    hue: int
    saturation: int
    value: int


class Light(Module, ABC):
    """Base class for TP-Link Light."""

    @property
    @abstractmethod
    def hsv(self) -> Annotated[HSV, FeatureAttribute()]:
        """Return the current HSV state of the bulb.

        :return: hue, saturation and value (degrees, %, %)
        """

    @property
    @abstractmethod
    def color_temp(self) -> Annotated[int, FeatureAttribute()]:
        """Whether the bulb supports color temperature changes."""

    @property
    @abstractmethod
    def brightness(self) -> Annotated[int, FeatureAttribute()]:
        """Return the current brightness in percentage."""

    @abstractmethod
    async def set_hsv(
        self,
        hue: int,
        saturation: int,
        value: int | None = None,
        *,
        transition: int | None = None,
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set new HSV.

        Note, transition is not supported and will be ignored.

        :param int hue: hue in degrees
        :param int saturation: saturation in percentage [0,100]
        :param int value: value in percentage [0, 100]
        :param int transition: transition in milliseconds.
        """

    @abstractmethod
    async def set_color_temp(
        self, temp: int, *, brightness: int | None = None, transition: int | None = None
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set the color temperature of the device in kelvin.

        Note, transition is not supported and will be ignored.

        :param int temp: The new color temperature, in Kelvin
        :param int transition: transition in milliseconds.
        """

    @abstractmethod
    async def set_brightness(
        self, brightness: int, *, transition: int | None = None
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set the brightness in percentage.

        Note, transition is not supported and will be ignored.

        :param int brightness: brightness in percent
        :param int transition: transition in milliseconds.
        """

    @property
    @abstractmethod
    def state(self) -> LightState:
        """Return the current light state."""

    @abstractmethod
    async def set_state(self, state: LightState) -> dict:
        """Set the light state."""

    def _deprecated_valid_temperature_range(self) -> ColorTempRange:
        if not (temp := self.get_feature("color_temp")):
            raise KasaException("Color temperature not supported")
        return ColorTempRange(temp.minimum_value, temp.maximum_value)

    def _deprecated_attributes(self, dep_name: str) -> str | None:
        map: dict[str, str] = {
            "is_color": "hsv",
            "is_dimmable": "brightness",
            "is_variable_color_temp": "color_temp",
        }
        return map.get(dep_name)

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            if name == "valid_temperature_range":
                msg = (
                    "valid_temperature_range is deprecated, use "
                    'get_feature("color_temp") minimum_value '
                    " and maximum_value instead"
                )
                warn(msg, DeprecationWarning, stacklevel=2)
                res = self._deprecated_valid_temperature_range()
                return res

            if name == "has_effects":
                msg = (
                    "has_effects is deprecated, check `Module.LightEffect "
                    "in device.modules` instead"
                )
                warn(msg, DeprecationWarning, stacklevel=2)
                return Module.LightEffect in self._device.modules

            if attr := self._deprecated_attributes(name):
                msg = f'{name} is deprecated, use has_feature("{attr}") instead'
                warn(msg, DeprecationWarning, stacklevel=2)
                return self.has_feature(attr)

            raise AttributeError(f"Energy module has no attribute {name!r}")



import pytest

from kasa.smart import SmartDevice

from ...conftest import variable_temp_smart


@variable_temp_smart
async def test_colortemp_component(dev: SmartDevice):
    """Test brightness feature."""
    assert isinstance(dev, SmartDevice)
    assert "color_temperature" in dev._components

    # Test getting the value
    feature = dev.features["color_temperature"]
    assert isinstance(feature.value, int)
    assert isinstance(feature.minimum_value, int)
    assert isinstance(feature.maximum_value, int)

    # Test setting the value
    # We need to take the min here, as L9xx reports a range [9000, 9000].
    new_value = min(feature.minimum_value + 1, feature.maximum_value)
    await feature.set_value(new_value)
    await dev.update()
    assert feature.value == new_value

    with pytest.raises(ValueError, match="out of range"):
        await feature.set_value(feature.minimum_value - 10)

    with pytest.raises(ValueError, match="out of range"):
        await feature.set_value(feature.maximum_value + 10)



import json
from dataclasses import replace
from json import dumps as json_dumps
from json import loads as json_loads

import aiohttp
import pytest
from mashumaro import MissingField

from kasa.credentials import Credentials
from kasa.deviceconfig import (
    DeviceConfig,
    DeviceConnectionParameters,
    DeviceEncryptionType,
    DeviceFamily,
)

from .conftest import load_fixture

PLUG_XOR_CONFIG = DeviceConfig(host="127.0.0.1")
PLUG_KLAP_CONFIG = DeviceConfig(
    host="127.0.0.1",
    connection_type=DeviceConnectionParameters(
        DeviceFamily.SmartTapoPlug, DeviceEncryptionType.Klap, login_version=2
    ),
)
CAMERA_AES_CONFIG = DeviceConfig(
    host="127.0.0.1",
    connection_type=DeviceConnectionParameters(
        DeviceFamily.SmartIpCamera, DeviceEncryptionType.Aes, https=True
    ),
)


async def test_serialization():
    """Test device config serialization."""
    config = DeviceConfig(host="Foo", http_client=aiohttp.ClientSession())
    config_dict = config.to_dict()
    config_json = json_dumps(config_dict)
    config2_dict = json_loads(config_json)
    config2 = DeviceConfig.from_dict(config2_dict)
    assert config == config2
    assert config.to_dict_control_credentials() == config.to_dict()


@pytest.mark.parametrize(
    ("fixture_name", "expected_value"),
    [
        ("deviceconfig_plug-xor.json", PLUG_XOR_CONFIG),
        ("deviceconfig_plug-klap.json", PLUG_KLAP_CONFIG),
        ("deviceconfig_camera-aes-https.json", CAMERA_AES_CONFIG),
    ],
    ids=lambda arg: arg.split("_")[-1] if isinstance(arg, str) else "",
)
async def test_deserialization(fixture_name: str, expected_value: DeviceConfig):
    """Test device config deserialization."""
    dict_val = json.loads(load_fixture("serialization", fixture_name))
    config = DeviceConfig.from_dict(dict_val)
    assert config == expected_value
    assert expected_value.to_dict() == dict_val


async def test_serialization_http_client():
    """Test that the http client does not try to serialize."""
    dict_val = json.loads(load_fixture("serialization", "deviceconfig_plug-klap.json"))

    config = replace(PLUG_KLAP_CONFIG, http_client=object())
    assert config.http_client

    assert config.to_dict() == dict_val


async def test_conn_param_no_https():
    """Test no https in connection param defaults to False."""
    dict_val = {
        "device_family": "SMART.TAPOPLUG",
        "encryption_type": "KLAP",
        "login_version": 2,
    }
    param = DeviceConnectionParameters.from_dict(dict_val)
    assert param.https is False
    assert param.to_dict() == {**dict_val, "https": False}


@pytest.mark.parametrize(
    ("input_value", "expected_error"),
    [
        ({"Foo": "Bar"}, MissingField),
        ("foobar", ValueError),
    ],
    ids=["invalid-dict", "not-dict"],
)
def test_deserialization_errors(input_value, expected_error):
    with pytest.raises(expected_error):
        DeviceConfig.from_dict(input_value)


async def test_credentials_hash():
    config = DeviceConfig(
        host="Foo",
        http_client=aiohttp.ClientSession(),
        credentials=Credentials("foo", "bar"),
    )
    config_dict = config.to_dict_control_credentials(credentials_hash="credhash")
    config_json = json_dumps(config_dict)
    config2_dict = json_loads(config_json)
    config2 = DeviceConfig.from_dict(config2_dict)
    assert config2.credentials_hash == "credhash"
    assert config2.credentials is None


async def test_blank_credentials_hash():
    config = DeviceConfig(
        host="Foo",
        http_client=aiohttp.ClientSession(),
        credentials=Credentials("foo", "bar"),
    )
    config_dict = config.to_dict_control_credentials(credentials_hash="")
    config_json = json_dumps(config_dict)
    config2_dict = json_loads(config_json)
    config2 = DeviceConfig.from_dict(config2_dict)
    assert config2.credentials_hash is None
    assert config2.credentials is None


async def test_exclude_credentials():
    config = DeviceConfig(
        host="Foo",
        http_client=aiohttp.ClientSession(),
        credentials=Credentials("foo", "bar"),
    )
    config_dict = config.to_dict_control_credentials(exclude_credentials=True)
    config_json = json_dumps(config_dict)
    config2_dict = json_loads(config_json)
    config2 = DeviceConfig.from_dict(config2_dict)
    assert config2.credentials is None



from __future__ import annotations

import pytest
from pytest_mock import MockerFixture

from kasa import Module
from kasa.smart import SmartDevice
from kasa.smart.modules import Alarm

from ...device_fixtures import get_parent_and_child_modules, parametrize

alarm = parametrize("has alarm", component_filter="alarm", protocol_filter={"SMART"})


@alarm
@pytest.mark.parametrize(
    ("feature", "prop_name", "type"),
    [
        ("alarm", "active", bool),
        ("alarm_source", "source", str | None),
        ("alarm_sound", "alarm_sound", str),
        ("alarm_volume", "_alarm_volume_str", str),
        ("alarm_volume_level", "alarm_volume", int),
    ],
)
async def test_features(dev: SmartDevice, feature: str, prop_name: str, type: type):
    """Test that features are registered and work as expected."""
    alarm = next(get_parent_and_child_modules(dev, Module.Alarm))
    assert alarm is not None

    prop = getattr(alarm, prop_name)
    assert isinstance(prop, type)

    feat = alarm._device.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)


@alarm
async def test_volume_feature(dev: SmartDevice):
    """Test that volume features have correct choices and range."""
    alarm = next(get_parent_and_child_modules(dev, Module.Alarm))
    assert alarm is not None

    volume_str_feat = alarm.get_feature("_alarm_volume_str")
    assert volume_str_feat

    assert volume_str_feat.choices == ["mute", "low", "normal", "high"]

    volume_int_feat = alarm.get_feature("alarm_volume")
    assert volume_int_feat.minimum_value == 0
    assert volume_int_feat.maximum_value == 3


@alarm
@pytest.mark.parametrize(
    ("kwargs", "request_params"),
    [
        pytest.param({"volume": "low"}, {"alarm_volume": "low"}, id="volume"),
        pytest.param({"volume": 0}, {"alarm_volume": "mute"}, id="volume-integer"),
        pytest.param({"duration": 1}, {"alarm_duration": 1}, id="duration"),
        pytest.param(
            {"sound": "Doorbell Ring 1"}, {"alarm_type": "Doorbell Ring 1"}, id="sound"
        ),
    ],
)
async def test_play(dev: SmartDevice, kwargs, request_params, mocker: MockerFixture):
    """Test that play parameters are handled correctly."""
    alarm: Alarm = next(get_parent_and_child_modules(dev, Module.Alarm))
    call_spy = mocker.spy(alarm, "call")
    await alarm.play(**kwargs)

    call_spy.assert_called_with("play_alarm", request_params)

    with pytest.raises(ValueError, match="Invalid duration"):
        await alarm.play(duration=-1)

    with pytest.raises(ValueError, match="Invalid sound"):
        await alarm.play(sound="unknown")

    with pytest.raises(ValueError, match="Invalid volume"):
        await alarm.play(volume="unknown")  # type: ignore[arg-type]

    with pytest.raises(ValueError, match="Invalid volume"):
        await alarm.play(volume=-1)


@alarm
async def test_stop(dev: SmartDevice, mocker: MockerFixture):
    """Test that stop creates the correct call."""
    alarm: Alarm = next(get_parent_and_child_modules(dev, Module.Alarm))
    call_spy = mocker.spy(alarm, "call")
    await alarm.stop()

    call_spy.assert_called_with("stop_alarm")


@alarm
@pytest.mark.parametrize(
    ("method", "value", "target_key"),
    [
        pytest.param(
            "set_alarm_sound", "Doorbell Ring 1", "type", id="set_alarm_sound"
        ),
        pytest.param("set_alarm_volume", "low", "volume", id="set_alarm_volume"),
        pytest.param("set_alarm_duration", 10, "duration", id="set_alarm_duration"),
    ],
)
async def test_set_alarm_configure(
    dev: SmartDevice,
    mocker: MockerFixture,
    method: str,
    value: str | int,
    target_key: str,
):
    """Test that set_alarm_sound creates the correct call."""
    alarm: Alarm = next(get_parent_and_child_modules(dev, Module.Alarm))
    call_spy = mocker.spy(alarm, "call")
    await getattr(alarm, method)(value)

    expected_params = {"duration": mocker.ANY, "type": mocker.ANY, "volume": mocker.ANY}
    expected_params[target_key] = value

    call_spy.assert_called_with("set_alarm_configure", expected_params)



"""Module for child devices."""

from ...device_type import DeviceType
from ..smartcammodule import SmartCamModule


class ChildDevice(SmartCamModule):
    """Implementation for child devices."""

    REQUIRED_COMPONENT = "childControl"
    NAME = "childdevice"
    QUERY_GETTER_NAME = "getChildDeviceList"
    # This module is unusual in that QUERY_MODULE_NAME in the response is not
    # the same one used in the request.
    QUERY_MODULE_NAME = "child_device_list"

    def query(self) -> dict:
        """Query to execute during the update cycle.

        Default implementation uses the raw query getter w/o parameters.
        """
        q = {self.QUERY_GETTER_NAME: {"childControl": {"start_index": 0}}}
        if self._device.device_type is DeviceType.Hub:
            q["getChildDeviceComponentList"] = {"childControl": {"start_index": 0}}
        return q

    async def _check_supported(self) -> bool:
        """Additional check to see if the module is supported by the device."""
        return self._device.device_type is DeviceType.Hub



from kasa import Device, Module

from ...device_fixtures import parametrize

triggerlogs = parametrize(
    "has trigger_logs",
    component_filter="trigger_log",
    protocol_filter={"SMART", "SMART.CHILD"},
)


@triggerlogs
async def test_trigger_logs(dev: Device):
    """Test that features are registered and work as expected."""
    triggerlogs = dev.modules.get(Module.TriggerLogs)
    assert triggerlogs is not None
    if logs := triggerlogs.logs:
        first = logs[0]
        assert isinstance(first.id, int)
        assert isinstance(first.timestamp, int)
        assert isinstance(first.event, str)
        assert isinstance(first.event_id, str)



"""Module for light strips (KL430)."""

from __future__ import annotations

from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..module import Module
from ..protocols import BaseProtocol
from .iotbulb import IotBulb
from .iotdevice import requires_update
from .modules.lighteffect import LightEffect


class IotLightStrip(IotBulb):
    """Representation of a TP-Link Smart light strip.

    Light strips work similarly to bulbs, but use a different service for controlling,
    and expose some extra information (such as length and active effect).
    This class extends :class:`SmartBulb` interface.

    Examples:
        >>> import asyncio
        >>> strip = IotLightStrip("127.0.0.1")
        >>> asyncio.run(strip.update())
        >>> print(strip.alias)
        Bedroom Lightstrip

        Getting the length of the strip:

        >>> strip.length
        16

        Currently active effect:

        >>> strip.effect
        {'brightness': 100, 'custom': 0, 'enable': 0,
        'id': 'bCTItKETDFfrKANolgldxfgOakaarARs', 'name': 'Flicker'}

    .. note::
        The device supports some features that are not currently implemented,
        feel free to find out how to control them and create a PR!


    See :class:`SmartBulb` for more examples.
    """

    LIGHT_SERVICE = "smartlife.iot.lightStrip"
    SET_LIGHT_METHOD = "set_light_state"

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.LightStrip

    async def _initialize_modules(self) -> None:
        """Initialize modules not added in init."""
        await super()._initialize_modules()
        self.add_module(
            Module.LightEffect,
            LightEffect(self, "smartlife.iot.lighting_effect"),
        )

    @property  # type: ignore
    @requires_update
    def length(self) -> int:
        """Return length of the strip."""
        return self.sys_info["length"]



"""Module for a SMART device."""

from __future__ import annotations

import base64
import logging
import time
from collections import OrderedDict
from collections.abc import Sequence
from datetime import UTC, datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from ..device import Device, DeviceInfo, WifiNetwork
from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..exceptions import AuthenticationError, DeviceError, KasaException, SmartErrorCode
from ..feature import Feature
from ..module import Module
from ..modulemapping import ModuleMapping, ModuleName
from ..protocols import SmartProtocol
from ..transports import AesTransport
from .modules import (
    ChildDevice,
    Cloud,
    DeviceModule,
    Firmware,
    Light,
    Thermostat,
    Time,
)
from .smartmodule import SmartModule

if TYPE_CHECKING:
    from .smartchilddevice import SmartChildDevice
_LOGGER = logging.getLogger(__name__)


# List of modules that non hub devices with children, i.e. ks240/P300, report on
# the child but only work on the parent.  See longer note below in _initialize_modules.
# This list should be updated when creating new modules that could have the
# same issue, homekit perhaps?
NON_HUB_PARENT_ONLY_MODULES = [DeviceModule, Time, Firmware, Cloud]

ComponentsRaw: TypeAlias = dict[str, list[dict[str, int | str]]]


# Device must go last as the other interfaces also inherit Device
# and python needs a consistent method resolution order.
class SmartDevice(Device):
    """Base class to represent a SMART protocol based device."""

    # Modules that are called as part of the init procedure on first update
    FIRST_UPDATE_MODULES = {DeviceModule, ChildDevice, Cloud}

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: SmartProtocol | None = None,
    ) -> None:
        _protocol = protocol or SmartProtocol(
            transport=AesTransport(config=config or DeviceConfig(host=host)),
        )
        super().__init__(host=host, config=config, protocol=_protocol)
        self.protocol: SmartProtocol
        self._components_raw: ComponentsRaw | None = None
        self._components: dict[str, int] = {}
        self._state_information: dict[str, Any] = {}
        self._modules: OrderedDict[str | ModuleName[Module], SmartModule] = (
            OrderedDict()
        )
        self._parent: SmartDevice | None = None
        self._children: dict[str, SmartDevice] = {}
        self._last_update_time: float | None = None
        self._on_since: datetime | None = None
        self._info: dict[str, Any] = {}
        self._logged_missing_child_ids: set[str] = set()

    async def _initialize_children(self) -> None:
        """Initialize children for power strips."""
        child_info_query = {
            "get_child_device_component_list": None,
            "get_child_device_list": None,
        }
        resp = await self.protocol.query(child_info_query)
        self.internal_state.update(resp)

    async def _try_create_child(
        self, info: dict, child_components: dict
    ) -> SmartDevice | None:
        from .smartchilddevice import SmartChildDevice

        return await SmartChildDevice.create(
            parent=self,
            child_info=info,
            child_components_raw=child_components,
        )

    async def _create_delete_children(
        self,
        child_device_resp: dict[str, list],
        child_device_components_resp: dict[str, list],
    ) -> bool:
        """Create and delete children. Return True if children changed.

        Adds newly found children and deletes children that are no longer
        reported by the device. It will only log once per child_id that
        can't be created to avoid spamming the logs on every update.
        """
        changed = False
        smart_children_components = {
            child["device_id"]: child
            for child in child_device_components_resp["child_component_list"]
        }
        children = self._children
        child_ids: set[str] = set()
        existing_child_ids = set(self._children.keys())

        for info in child_device_resp["child_device_list"]:
            if (child_id := info.get("device_id")) and (
                child_components := smart_children_components.get(child_id)
            ):
                child_ids.add(child_id)

                if child_id in existing_child_ids:
                    continue

                child = await self._try_create_child(info, child_components)
                if child:
                    _LOGGER.debug("Created child device %s for %s", child, self.host)
                    changed = True
                    children[child_id] = child
                    continue

                if child_id not in self._logged_missing_child_ids:
                    self._logged_missing_child_ids.add(child_id)
                    _LOGGER.debug("Child device type not supported: %s", info)
                continue

            if child_id:
                if child_id not in self._logged_missing_child_ids:
                    self._logged_missing_child_ids.add(child_id)
                    _LOGGER.debug(
                        "Could not find child components for device %s, "
                        "child_id %s, components: %s: ",
                        self.host,
                        child_id,
                        smart_children_components,
                    )
                continue

            # If we couldn't get a child device id we still only want to
            # log once to avoid spamming the logs on every update cycle
            # so store it under an empty string
            if "" not in self._logged_missing_child_ids:
                self._logged_missing_child_ids.add("")
                _LOGGER.debug(
                    "Could not find child id for device %s, info: %s", self.host, info
                )

        removed_ids = existing_child_ids - child_ids
        for removed_id in removed_ids:
            changed = True
            removed = children.pop(removed_id)
            _LOGGER.debug("Removed child device %s from %s", removed, self.host)

        return changed

    @property
    def children(self) -> Sequence[SmartDevice]:
        """Return list of children."""
        return list(self._children.values())

    @property
    def modules(self) -> ModuleMapping[SmartModule]:
        """Return the device modules."""
        return cast(ModuleMapping[SmartModule], self._modules)

    def _try_get_response(
        self, responses: dict, request: str, default: Any | None = None
    ) -> dict:
        response = responses.get(request)
        if isinstance(response, SmartErrorCode):
            _LOGGER.debug(
                "Error %s getting request %s for device %s",
                response,
                request,
                self.host,
            )
            response = None
        if response is not None:
            return response
        if default is not None:
            return default
        raise KasaException(
            f"{request} not found in {responses} for device {self.host}"
        )

    @staticmethod
    def _parse_components(components_raw: ComponentsRaw) -> dict[str, int]:
        return {
            str(comp["id"]): int(comp["ver_code"])
            for comp in components_raw["component_list"]
        }

    async def _negotiate(self) -> None:
        """Perform initialization.

        We fetch the device info and the available components as early as possible.
        If the device reports supporting child devices, they are also initialized.
        """
        initial_query = {
            "component_nego": None,
            "get_device_info": None,
            "get_connect_cloud_state": None,
        }
        resp = await self.protocol.query(initial_query)

        # Save the initial state to allow modules access the device info already
        # during the initialization, which is necessary as some information like the
        # supported color temperature range is contained within the response.
        self._last_update.update(resp)
        self._info = self._try_get_response(resp, "get_device_info")

        # Create our internal presentation of available components
        self._components_raw = cast(ComponentsRaw, resp["component_nego"])

        self._components = self._parse_components(self._components_raw)

        if "child_device" in self._components and not self.children:
            await self._initialize_children()

    async def _update_children_info(self) -> bool:
        """Update the internal child device info from the parent info.

        Return true if children added or deleted.
        """
        changed = False
        if child_info := self._try_get_response(
            self._last_update, "get_child_device_list", {}
        ):
            changed = await self._create_delete_children(
                child_info, self._last_update["get_child_device_component_list"]
            )

            for info in child_info["child_device_list"]:
                child_id = info.get("device_id")
                if child_id not in self._children:
                    # _create_delete_children has already logged a message
                    continue

                self._children[child_id]._update_internal_state(info)

        return changed

    def _update_internal_info(self, info_resp: dict) -> None:
        """Update the internal device info."""
        self._info = self._try_get_response(info_resp, "get_device_info")

    async def update(self, update_children: bool = True) -> None:
        """Update the device."""
        if self.credentials is None and self.credentials_hash is None:
            raise AuthenticationError("Tapo plug requires authentication.")

        first_update = self._last_update_time is None
        now = time.monotonic()
        self._last_update_time = now

        if first_update:
            await self._negotiate()
            await self._initialize_modules()
            # Run post update for the cloud module
            if cloud_mod := self.modules.get(Module.Cloud):
                await self._handle_module_post_update(cloud_mod, now, had_query=True)

        resp = await self._modular_update(first_update, now)

        children_changed = await self._update_children_info()
        # Call child update which will only update module calls, info is updated
        # from get_child_device_list. update_children only affects hub devices, other
        # devices will always update children to prevent errors on module access.
        # This needs to go after updating the internal state of the children so that
        # child modules have access to their sysinfo.
        if children_changed or update_children or self.device_type != DeviceType.Hub:
            for child in self._children.values():
                if TYPE_CHECKING:
                    assert isinstance(child, SmartChildDevice)
                await child._update()

        # We can first initialize the features after the first update.
        # We make here an assumption that every device has at least a single feature.
        if not self._features:
            await self._initialize_features()

        if _LOGGER.isEnabledFor(logging.DEBUG):
            updated = self._last_update if first_update else resp
            _LOGGER.debug("Update completed %s: %s", self.host, list(updated.keys()))

    async def _handle_module_post_update(
        self, module: SmartModule, update_time: float, had_query: bool
    ) -> None:
        if module.disabled:
            return  # pragma: no cover
        if had_query:
            module._last_update_time = update_time
        try:
            await module._post_update_hook()
            module._set_error(None)
        except Exception as ex:
            # Only set the error if a query happened.
            if had_query:
                module._set_error(ex)
                _LOGGER.warning(
                    "Error processing %s for device %s, module will be unavailable: %s",
                    module.name,
                    self.host,
                    ex,
                )

    async def _modular_update(
        self, first_update: bool, update_time: float
    ) -> dict[str, Any]:
        """Update the device with via the module queries."""
        req: dict[str, Any] = {}
        # Keep a track of actual module queries so we can track the time for
        # modules that do not need to be updated frequently
        module_queries: list[SmartModule] = []
        mq = {
            module: query
            for module in self._modules.values()
            if (first_update or module.disabled is False) and (query := module.query())
        }
        for module, query in mq.items():
            if first_update and module.__class__ in self.FIRST_UPDATE_MODULES:
                module._last_update_time = update_time
                continue
            if module._should_update(update_time):
                module_queries.append(module)
                req.update(query)

        _LOGGER.debug(
            "Querying %s for modules: %s",
            self.host,
            ", ".join(mod.name for mod in module_queries),
        )

        try:
            resp = await self.protocol.query(req)
        except Exception as ex:
            resp = await self._handle_modular_update_error(
                ex, first_update, ", ".join(mod.name for mod in module_queries), req
            )

        info_resp = self._last_update if first_update else resp
        self._last_update.update(**resp)
        self._update_internal_info(info_resp)

        # Call handle update for modules that want to update internal data
        for module in self._modules.values():
            await self._handle_module_post_update(
                module, update_time, had_query=module in module_queries
            )

        return resp

    async def _handle_modular_update_error(
        self,
        ex: Exception,
        first_update: bool,
        module_names: str,
        requests: dict[str, Any],
    ) -> dict[str, Any]:
        """Handle an error on calling module update.

        Will try to call all modules individually
        and any errors such as timeouts will be set as a SmartErrorCode.
        """
        msg_part = "on first update" if first_update else "after first update"

        _LOGGER.error(
            "Error querying %s for modules '%s' %s: %s",
            self.host,
            module_names,
            msg_part,
            ex,
        )
        responses = {}
        for meth, params in requests.items():
            try:
                resp = await self.protocol.query({meth: params})
                responses[meth] = resp[meth]
            except Exception as iex:
                _LOGGER.error(
                    "Error querying %s individually for module query '%s' %s: %s",
                    self.host,
                    meth,
                    msg_part,
                    iex,
                )
                responses[meth] = SmartErrorCode.INTERNAL_QUERY_ERROR
        return responses

    async def _initialize_modules(self) -> None:
        """Initialize modules based on component negotiation response."""
        from .smartmodule import SmartModule

        # Some wall switches (like ks240) are internally presented as having child
        # devices which report the child's components on the parent's sysinfo, even
        # when they need to be accessed through the children.
        # The logic below ensures that such devices add all but whitelisted, only on
        # the child device.
        # It also ensures that devices like power strips do not add modules such as
        # firmware to the child devices.
        skip_parent_only_modules = False
        child_modules_to_skip: dict = {}  # TODO: this is never non-empty
        if self._parent and self._parent.device_type != DeviceType.Hub:
            skip_parent_only_modules = True

        for mod in SmartModule.REGISTERED_MODULES.values():
            if (
                skip_parent_only_modules and mod in NON_HUB_PARENT_ONLY_MODULES
            ) or mod.__name__ in child_modules_to_skip:
                continue
            required_component = cast(str, mod.REQUIRED_COMPONENT)
            if required_component in self._components or any(
                self.sys_info.get(key) is not None for key in mod.SYSINFO_LOOKUP_KEYS
            ):
                _LOGGER.debug(
                    "Device %s, found required %s, adding %s to modules.",
                    self.host,
                    required_component,
                    mod.__name__,
                )
                module = mod(self, required_component)
                if await module._check_supported():
                    self._modules[module.name] = module

        if (
            Module.Brightness in self._modules
            or Module.Color in self._modules
            or Module.ColorTemperature in self._modules
        ):
            self._modules[Light.__name__] = Light(self, "light")
        if (
            Module.TemperatureControl in self._modules
            and Module.TemperatureSensor in self._modules
        ):
            self._modules[Thermostat.__name__] = Thermostat(self, "thermostat")

        # We move time to the beginning so other modules can access the
        # time and timezone after update if required. e.g. cleanrecords
        if Time.__name__ in self._modules:
            self._modules.move_to_end(Time.__name__, last=False)

    async def _initialize_features(self) -> None:
        """Initialize device features."""
        self._add_feature(
            Feature(
                self,
                id="device_id",
                name="Device ID",
                attribute_getter="device_id",
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )
        if "device_on" in self._info:
            self._add_feature(
                Feature(
                    self,
                    id="state",
                    name="State",
                    attribute_getter="is_on",
                    attribute_setter="set_state",
                    type=Feature.Type.Switch,
                    category=Feature.Category.Primary,
                )
            )

        if "signal_level" in self._info:
            self._add_feature(
                Feature(
                    self,
                    id="signal_level",
                    name="Signal Level",
                    attribute_getter=lambda x: x._info["signal_level"],
                    icon="mdi:signal",
                    category=Feature.Category.Info,
                    type=Feature.Type.Sensor,
                )
            )

        if "rssi" in self._info:
            self._add_feature(
                Feature(
                    self,
                    id="rssi",
                    name="RSSI",
                    attribute_getter=lambda x: x._info["rssi"],
                    icon="mdi:signal",
                    unit_getter=lambda: "dBm",
                    category=Feature.Category.Debug,
                    type=Feature.Type.Sensor,
                )
            )

        if "ssid" in self._info:
            self._add_feature(
                Feature(
                    device=self,
                    id="ssid",
                    name="SSID",
                    attribute_getter="ssid",
                    icon="mdi:wifi",
                    category=Feature.Category.Debug,
                    type=Feature.Type.Sensor,
                )
            )

        # We check for the key available, and not for the property truthiness,
        # as the value is falsy when the device is off.
        if "on_time" in self._info:
            self._add_feature(
                Feature(
                    device=self,
                    id="on_since",
                    name="On since",
                    attribute_getter="on_since",
                    icon="mdi:clock",
                    category=Feature.Category.Debug,
                    type=Feature.Type.Sensor,
                )
            )

        self._add_feature(
            Feature(
                device=self,
                id="reboot",
                name="Reboot",
                attribute_setter="reboot",
                icon="mdi:restart",
                category=Feature.Category.Debug,
                type=Feature.Type.Action,
            )
        )

        if self.parent is not None and (
            cs := self.parent.modules.get(Module.ChildSetup)
        ):
            self._add_feature(
                Feature(
                    device=self,
                    id="unpair",
                    name="Unpair device",
                    container=cs,
                    attribute_setter=lambda: cs.unpair(self.device_id),
                    category=Feature.Category.Debug,
                    type=Feature.Type.Action,
                )
            )

        for module in self.modules.values():
            module._initialize_features()
            for feat in module._module_features.values():
                self._add_feature(feat)

    @property
    def _is_hub_child(self) -> bool:
        """Returns true if the device is a child of a hub."""
        return self.parent is not None and self.parent.device_type is DeviceType.Hub

    @property
    def is_cloud_connected(self) -> bool:
        """Returns if the device is connected to the cloud."""
        if Module.Cloud not in self.modules:
            return False
        return self.modules[Module.Cloud].is_connected

    @property
    def sys_info(self) -> dict[str, Any]:
        """Returns the device info."""
        return self._info  # type: ignore

    @property
    def model(self) -> str:
        """Returns the device model."""
        # If update hasn't been called self._device_info can't be used
        if self._last_update:
            return self.device_info.short_name

        disco_model = str(self._info.get("device_model"))
        long_name, _, _ = disco_model.partition("(")
        return long_name

    @property
    def alias(self) -> str | None:
        """Returns the device alias or nickname."""
        if self._info and (nickname := self._info.get("nickname")):
            return base64.b64decode(nickname).decode()
        else:
            return None

    @property
    def time(self) -> datetime:
        """Return the time."""
        if (time_mod := self.modules.get(Module.Time)) or (
            self._parent and (time_mod := self._parent.modules.get(Module.Time))
        ):
            return time_mod.time

        # We have no device time, use current local time.
        return datetime.now(UTC).astimezone().replace(microsecond=0)

    @property
    def on_since(self) -> datetime | None:
        """Return the time that the device was turned on or None if turned off.

        This returns a cached value if the device reported value difference is under
        five seconds to avoid device-caused jitter.
        """
        if (
            not self._info.get("device_on")
            or (on_time := self._info.get("on_time")) is None
        ):
            self._on_since = None
            return None

        on_time = cast(float, on_time)
        on_since = self.time - timedelta(seconds=on_time)
        if not self._on_since or timedelta(
            seconds=0
        ) < on_since - self._on_since > timedelta(seconds=5):
            self._on_since = on_since
        return self._on_since

    @property
    def timezone(self) -> tzinfo:
        """Return the timezone and time_difference."""
        if TYPE_CHECKING:
            assert self.time.tzinfo
        return self.time.tzinfo

    @property
    def hw_info(self) -> dict:
        """Return hardware info for the device."""
        return {
            "sw_ver": self._info.get("fw_ver"),
            "hw_ver": self._info.get("hw_ver"),
            "mac": self._info.get("mac"),
            "type": self._info.get("type"),
            "hwId": self._info.get("device_id"),
            "dev_name": self.alias,
            "oemId": self._info.get("oem_id"),
        }

    @property
    def location(self) -> dict:
        """Return the device location."""
        loc = {
            "latitude": cast(float, self._info.get("latitude", 0)) / 10_000,
            "longitude": cast(float, self._info.get("longitude", 0)) / 10_000,
        }
        return loc

    @property
    def rssi(self) -> int | None:
        """Return the rssi."""
        rssi = self._info.get("rssi")
        return int(rssi) if rssi else None

    @property
    def mac(self) -> str:
        """Return the mac formatted with colons."""
        return str(self._info.get("mac")).replace("-", ":")

    @property
    def device_id(self) -> str:
        """Return the device id."""
        return str(self._info.get("device_id"))

    @property
    def internal_state(self) -> dict:
        """Return all the internal state data."""
        return self._last_update

    def _update_internal_state(self, info: dict[str, Any]) -> None:
        """Update the internal info state.

        This is used by the parent to push updates to its children.
        """
        self._info = info

    async def _query_helper(self, method: str, params: dict | None = None) -> dict:
        return await self.protocol.query({method: params})

    @property
    def ssid(self) -> str:
        """Return ssid of the connected wifi ap."""
        ssid = self._info.get("ssid")
        ssid = base64.b64decode(ssid).decode() if ssid else "No SSID"
        return ssid

    @property
    def has_emeter(self) -> bool:
        """Return if the device has emeter."""
        return Module.Energy in self.modules

    @property
    def is_on(self) -> bool:
        """Return true if the device is on."""
        return bool(self._info.get("device_on"))

    async def set_state(self, on: bool) -> dict:
        """Set the device state.

        See :meth:`is_on`.
        """
        return await self.protocol.query({"set_device_info": {"device_on": on}})

    async def turn_on(self, **kwargs: Any) -> dict:
        """Turn on the device."""
        return await self.set_state(True)

    async def turn_off(self, **kwargs: Any) -> dict:
        """Turn off the device."""
        return await self.set_state(False)

    def update_from_discover_info(
        self,
        info: dict,
    ) -> None:
        """Update state from info from the discover call."""
        self._discovery_info = info
        self._info = info

    async def wifi_scan(self) -> list[WifiNetwork]:
        """Scan for available wifi networks."""

        def _net_for_scan_info(res: dict) -> WifiNetwork:
            return WifiNetwork(
                ssid=base64.b64decode(res["ssid"]).decode(),
                cipher_type=res["cipher_type"],
                key_type=res["key_type"],
                channel=res["channel"],
                signal_level=res["signal_level"],
                bssid=res["bssid"],
            )

        _LOGGER.debug("Querying networks")

        resp = await self.protocol.query({"get_wireless_scan_info": {"start_index": 0}})
        networks = [
            _net_for_scan_info(net) for net in resp["get_wireless_scan_info"]["ap_list"]
        ]
        return networks

    async def wifi_join(
        self, ssid: str, password: str, keytype: str = "wpa2_psk"
    ) -> dict:
        """Join the given wifi network.

        This method returns nothing as the device tries to activate the new
        settings immediately instead of responding to the request.

        If joining the network fails, the device will return to the previous state
        after some delay.
        """
        if not self.credentials:
            raise AuthenticationError("Device requires authentication.")

        payload = {
            "account": {
                "username": base64.b64encode(
                    self.credentials.username.encode()
                ).decode(),
                "password": base64.b64encode(
                    self.credentials.password.encode()
                ).decode(),
            },
            "wireless": {
                "key_type": keytype,
                "password": base64.b64encode(password.encode()).decode(),
                "ssid": base64.b64encode(ssid.encode()).decode(),
            },
            "time": self.internal_state["get_device_time"],
        }

        # The device does not respond to the request but changes the settings
        # immediately which causes us to timeout.
        # Thus, We limit retries and suppress the raised exception as useless.
        try:
            return await self.protocol.query({"set_qs_info": payload}, retry_count=0)
        except DeviceError:
            raise  # Re-raise on device-reported errors
        except KasaException:
            _LOGGER.debug(
                "Received a kasa exception for wifi join, but this is expected"
            )
            return {}

    async def update_credentials(self, username: str, password: str) -> dict:
        """Update device credentials.

        This will replace the existing authentication credentials on the device.
        """
        time_data = self.internal_state["get_device_time"]
        payload = {
            "account": {
                "username": base64.b64encode(username.encode()).decode(),
                "password": base64.b64encode(password.encode()).decode(),
            },
            "time": time_data,
        }
        return await self.protocol.query({"set_qs_info": payload})

    async def set_alias(self, alias: str) -> dict:
        """Set the device name (alias)."""
        return await self.protocol.query(
            {"set_device_info": {"nickname": base64.b64encode(alias.encode()).decode()}}
        )

    async def reboot(self, delay: int = 1) -> None:
        """Reboot the device.

        Note that giving a delay of zero causes this to block,
        as the device reboots immediately without responding to the call.
        """
        await self.protocol.query({"device_reboot": {"delay": delay}})

    async def factory_reset(self) -> None:
        """Reset device back to factory settings.

        Note, this does not downgrade the firmware.
        """
        await self.protocol.query("device_reset")

    @property
    def device_type(self) -> DeviceType:
        """Return the device type."""
        if self._device_type is not DeviceType.Unknown:
            return self._device_type

        if (
            not (type_str := self._info.get("type", self._info.get("device_type")))
            or not self._components
        ):
            # no update or discovery info
            return self._device_type

        self._device_type = self._get_device_type_from_components(
            list(self._components.keys()), type_str
        )

        return self._device_type

    @staticmethod
    def _get_device_type_from_components(
        components: list[str], device_type: str
    ) -> DeviceType:
        """Find type to be displayed as a supported device category."""
        if "HUB" in device_type:
            return DeviceType.Hub
        if "PLUG" in device_type:
            if "child_device" in components:
                return DeviceType.Strip
            return DeviceType.Plug
        if "light_strip" in components:
            return DeviceType.LightStrip
        if "SWITCH" in device_type and "child_device" in components:
            return DeviceType.WallSwitch
        if "dimmer_calibration" in components:
            return DeviceType.Dimmer
        if "brightness" in components:
            return DeviceType.Bulb
        if "SWITCH" in device_type:
            return DeviceType.WallSwitch
        if "SENSOR" in device_type:
            return DeviceType.Sensor
        if "ENERGY" in device_type:
            return DeviceType.Thermostat
        if "ROBOVAC" in device_type:
            return DeviceType.Vacuum
        if "TAPOCHIME" in device_type:
            return DeviceType.Chime
        _LOGGER.warning("Unknown device type, falling back to plug")
        return DeviceType.Plug

    @staticmethod
    def _get_device_info(
        info: dict[str, Any], discovery_info: dict[str, Any] | None
    ) -> DeviceInfo:
        """Get model information for a device."""
        di = info["get_device_info"]
        components = [comp["id"] for comp in info["component_nego"]["component_list"]]

        # Get model/region info
        short_name = di["model"]
        region = None
        if discovery_info:
            device_model = discovery_info["device_model"]
            long_name, _, region = device_model.partition("(")
            if region:  # P100 doesn't have region
                region = region.replace(")", "")
        else:
            long_name = short_name
        if not region:  # some devices have region in specs
            region = di.get("specs")

        # Get other info
        device_family = di["type"]
        device_type = SmartDevice._get_device_type_from_components(
            components, device_family
        )
        fw_version_full = di["fw_ver"]
        if " " in fw_version_full:
            firmware_version, firmware_build = fw_version_full.split(" ", maxsplit=1)
        else:
            firmware_version, firmware_build = fw_version_full, None
        _protocol, devicetype = device_family.split(".")
        # Brand inferred from SMART.KASAPLUG/SMART.TAPOPLUG etc.
        brand = devicetype[:4].lower()

        return DeviceInfo(
            short_name=short_name,
            long_name=long_name,
            brand=brand,
            device_family=device_family,
            device_type=device_type,
            hardware_version=di["hw_ver"],
            firmware_version=firmware_version,
            firmware_build=firmware_build,
            requires_auth=True,
            region=region,
        )



"""Implementation of the TP-Link AES transport.

Based on the work of https://github.com/petretiandrea/plugp100
under compatible GNU GPL3 license.
"""

from __future__ import annotations

import base64
import hashlib
import logging
import time
from collections.abc import AsyncGenerator
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, cast

from cryptography.hazmat.primitives import hashes, padding, serialization
from cryptography.hazmat.primitives.asymmetric import padding as asymmetric_padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from yarl import URL

from kasa.credentials import DEFAULT_CREDENTIALS, Credentials, get_default_credentials
from kasa.deviceconfig import DeviceConfig
from kasa.exceptions import (
    SMART_AUTHENTICATION_ERRORS,
    SMART_RETRYABLE_ERRORS,
    AuthenticationError,
    DeviceError,
    KasaException,
    SmartErrorCode,
    TimeoutError,
    _ConnectionError,
    _RetryableError,
)
from kasa.httpclient import HttpClient
from kasa.json import dumps as json_dumps
from kasa.json import loads as json_loads

from .basetransport import BaseTransport

_LOGGER = logging.getLogger(__name__)


ONE_DAY_SECONDS = 86400
SESSION_EXPIRE_BUFFER_SECONDS = 60 * 20


def _sha1(payload: bytes) -> str:
    sha1_algo = hashlib.sha1()  # noqa: S324
    sha1_algo.update(payload)
    return sha1_algo.hexdigest()


class TransportState(Enum):
    """Enum for AES state."""

    HANDSHAKE_REQUIRED = auto()  # Handshake needed
    LOGIN_REQUIRED = auto()  # Login needed
    ESTABLISHED = auto()  # Ready to send requests


class AesTransport(BaseTransport):
    """Implementation of the AES encryption protocol.

    AES is the name used in device discovery for TP-Link's TAPO encryption
    protocol, sometimes used by newer firmware versions on kasa devices.
    """

    DEFAULT_PORT: int = 80
    SESSION_COOKIE_NAME = "TP_SESSIONID"
    TIMEOUT_COOKIE_NAME = "TIMEOUT"
    COMMON_HEADERS = {
        "Content-Type": "application/json",
        "requestByApp": "true",
        "Accept": "application/json",
    }
    CONTENT_LENGTH = "Content-Length"
    KEY_PAIR_CONTENT_LENGTH = 314

    def __init__(
        self,
        *,
        config: DeviceConfig,
    ) -> None:
        super().__init__(config=config)

        self._login_version = config.connection_type.login_version
        if (
            not self._credentials or self._credentials.username is None
        ) and not self._credentials_hash:
            self._credentials = Credentials()
        if self._credentials:
            self._login_params = self._get_login_params(self._credentials)
        else:
            self._login_params = json_loads(
                base64.b64decode(self._credentials_hash.encode()).decode()  # type: ignore[union-attr]
            )
        self._default_credentials: Credentials | None = None
        self._http_client: HttpClient = HttpClient(config)

        self._state = TransportState.HANDSHAKE_REQUIRED

        self._encryption_session: AesEncyptionSession | None = None
        self._session_expire_at: float | None = None

        self._session_cookie: dict[str, str] | None = None

        self._key_pair: KeyPair | None = None
        if config.aes_keys:
            aes_keys = config.aes_keys
            self._key_pair = KeyPair.create_from_der_keys(
                aes_keys["private"], aes_keys["public"]
            )
        self._app_url = URL(f"http://{self._host}:{self._port}/app")
        self._token_url: URL | None = None

        _LOGGER.debug("Created AES transport for %s", self._host)

    @property
    def default_port(self) -> int:
        """Default port for the transport."""
        if port := self._config.connection_type.http_port:
            return port
        return self.DEFAULT_PORT

    @property
    def credentials_hash(self) -> str | None:
        """The hashed credentials used by the transport."""
        if self._credentials == Credentials():
            return None
        return base64.b64encode(json_dumps(self._login_params).encode()).decode()

    def _get_login_params(self, credentials: Credentials) -> dict[str, str]:
        """Get the login parameters based on the login_version."""
        un, pw = self.hash_credentials(self._login_version == 2, credentials)
        password_field_name = "password2" if self._login_version == 2 else "password"
        return {password_field_name: pw, "username": un}

    @staticmethod
    def hash_credentials(login_v2: bool, credentials: Credentials) -> tuple[str, str]:
        """Hash the credentials."""
        un = base64.b64encode(_sha1(credentials.username.encode()).encode()).decode()
        if login_v2:
            pw = base64.b64encode(
                _sha1(credentials.password.encode()).encode()
            ).decode()
        else:
            pw = base64.b64encode(credentials.password.encode()).decode()
        return un, pw

    def _handle_response_error_code(self, resp_dict: dict, msg: str) -> None:
        error_code_raw = resp_dict.get("error_code")
        try:
            error_code = SmartErrorCode.from_int(error_code_raw)
        except ValueError:
            _LOGGER.warning(
                "Device %s received unknown error code: %s", self._host, error_code_raw
            )
            error_code = SmartErrorCode.INTERNAL_UNKNOWN_ERROR
        if error_code is SmartErrorCode.SUCCESS:
            return
        msg = f"{msg}: {self._host}: {error_code.name}({error_code.value})"
        if error_code in SMART_RETRYABLE_ERRORS:
            raise _RetryableError(msg, error_code=error_code)
        if error_code in SMART_AUTHENTICATION_ERRORS:
            self._state = TransportState.HANDSHAKE_REQUIRED
            raise AuthenticationError(msg, error_code=error_code)
        raise DeviceError(msg, error_code=error_code)

    async def send_secure_passthrough(self, request: str) -> dict[str, Any]:
        """Send encrypted message as passthrough."""
        if self._state is TransportState.ESTABLISHED and self._token_url:
            url = self._token_url
        else:
            url = self._app_url

        encrypted_payload = self._encryption_session.encrypt(request.encode())  # type: ignore
        passthrough_request = {
            "method": "securePassthrough",
            "params": {"request": encrypted_payload.decode()},
        }
        status_code, resp_dict = await self._http_client.post(
            url,
            json=passthrough_request,
            headers=self.COMMON_HEADERS,
            cookies_dict=self._session_cookie,
        )
        # _LOGGER.debug(f"secure_passthrough response is {status_code}: {resp_dict}")

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to passthrough"
            )

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)
            assert self._encryption_session is not None

        self._handle_response_error_code(
            resp_dict, "Error sending secure_passthrough message"
        )

        raw_response: str = resp_dict["result"]["response"]

        try:
            response = self._encryption_session.decrypt(raw_response.encode())
            ret_val = json_loads(response)
        except Exception as ex:
            try:
                ret_val = json_loads(raw_response)
                _LOGGER.debug(
                    "Received unencrypted response over secure passthrough from %s",
                    self._host,
                )
            except Exception:
                raise KasaException(
                    f"Unable to decrypt response from {self._host}, "
                    + f"error: {ex}, response: {raw_response}",
                    ex,
                ) from ex
        return ret_val  # type: ignore[return-value]

    async def perform_login(self) -> None:
        """Login to the device."""
        try:
            await self.try_login(self._login_params)
            _LOGGER.debug(
                "%s: logged in with provided credentials",
                self._host,
            )
        except AuthenticationError as aex:
            try:
                if aex.error_code is not SmartErrorCode.LOGIN_ERROR:
                    raise aex
                _LOGGER.debug(
                    "%s: trying login with default TAPO credentials",
                    self._host,
                )
                if self._default_credentials is None:
                    self._default_credentials = get_default_credentials(
                        DEFAULT_CREDENTIALS["TAPO"]
                    )
                await self.perform_handshake()
                await self.try_login(self._get_login_params(self._default_credentials))
                _LOGGER.debug(
                    "%s: logged in with default TAPO credentials",
                    self._host,
                )
            except (AuthenticationError, _ConnectionError, TimeoutError):
                raise
            except Exception as ex:
                raise KasaException(
                    "Unable to login and trying default "
                    + f"login raised another exception: {ex}",
                    ex,
                ) from ex

    async def try_login(self, login_params: dict[str, Any]) -> None:
        """Try to login with supplied login_params."""
        login_request = {
            "method": "login_device",
            "params": login_params,
            "request_time_milis": round(time.time() * 1000),
        }
        request = json_dumps(login_request)

        resp_dict = await self.send_secure_passthrough(request)
        self._handle_response_error_code(resp_dict, "Error logging in")
        login_token = resp_dict["result"]["token"]
        self._token_url = self._app_url.with_query(f"token={login_token}")
        self._state = TransportState.ESTABLISHED

    async def _generate_key_pair_payload(self) -> AsyncGenerator:
        """Generate the request body and return an ascyn_generator.

        This prevents the key pair being generated unless a connection
        can be made to the device.
        """
        _LOGGER.debug("Generating keypair")
        if not self._key_pair:
            kp = KeyPair.create_key_pair()
            self._config.aes_keys = {
                "private": kp.private_key_der_b64,
                "public": kp.public_key_der_b64,
            }
            self._key_pair = kp

        pub_key = (
            "-----BEGIN PUBLIC KEY-----\n"
            + self._key_pair.public_key_der_b64  # type: ignore[union-attr]
            + "\n-----END PUBLIC KEY-----\n"
        )
        handshake_params = {"key": pub_key}
        request_body = {"method": "handshake", "params": handshake_params}
        _LOGGER.debug("Handshake request: %s", request_body)
        yield json_dumps(request_body).encode()

    async def perform_handshake(self) -> None:
        """Perform the handshake."""
        _LOGGER.debug("Will perform handshaking...")

        self._token_url = None
        self._session_expire_at = None
        self._session_cookie = None

        # Device needs the content length or it will response with 500
        headers = {
            **self.COMMON_HEADERS,
            self.CONTENT_LENGTH: str(self.KEY_PAIR_CONTENT_LENGTH),
        }
        http_client = self._http_client

        status_code, resp_dict = await http_client.post(
            self._app_url,
            json=self._generate_key_pair_payload(),
            headers=headers,
            cookies_dict=self._session_cookie,
        )

        _LOGGER.debug("Device responded with: %s", resp_dict)

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to handshake"
            )

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)

        self._handle_response_error_code(resp_dict, "Unable to complete handshake")

        handshake_key = resp_dict["result"]["key"]

        if (
            cookie := http_client.get_cookie(self.SESSION_COOKIE_NAME)  # type: ignore
        ) or (
            cookie := http_client.get_cookie("SESSIONID")  # type: ignore
        ):
            self._session_cookie = {self.SESSION_COOKIE_NAME: cookie}

        timeout = int(
            http_client.get_cookie(self.TIMEOUT_COOKIE_NAME) or ONE_DAY_SECONDS
        )
        # There is a 24 hour timeout on the session cookie
        # but the clock on the device is not always accurate
        # so we set the expiry to 24 hours from now minus a buffer
        self._session_expire_at = time.time() + timeout - SESSION_EXPIRE_BUFFER_SECONDS
        if TYPE_CHECKING:
            assert self._key_pair is not None
        self._encryption_session = AesEncyptionSession.create_from_keypair(
            handshake_key, self._key_pair
        )

        self._state = TransportState.LOGIN_REQUIRED

        _LOGGER.debug("Handshake with %s complete", self._host)

    def _handshake_session_expired(self) -> bool:
        """Return true if session has expired."""
        return (
            self._session_expire_at is None
            or self._session_expire_at - time.time() <= 0
        )

    async def send(self, request: str) -> dict[str, Any]:
        """Send the request."""
        if (
            self._state is TransportState.HANDSHAKE_REQUIRED
            or self._handshake_session_expired()
        ):
            await self.perform_handshake()
        if self._state is not TransportState.ESTABLISHED:
            try:
                await self.perform_login()
            # After a login failure handshake needs to
            # be redone or a 9999 error is received.
            except AuthenticationError as ex:
                self._state = TransportState.HANDSHAKE_REQUIRED
                raise ex

        return await self.send_secure_passthrough(request)

    async def close(self) -> None:
        """Close the http client and reset internal state."""
        await self.reset()
        await self._http_client.close()

    async def reset(self) -> None:
        """Reset internal handshake and login state."""
        self._state = TransportState.HANDSHAKE_REQUIRED


class AesEncyptionSession:
    """Class for an AES encryption session."""

    @staticmethod
    def create_from_keypair(
        handshake_key: str, keypair: KeyPair
    ) -> AesEncyptionSession:
        """Create the encryption session."""
        handshake_key_bytes: bytes = base64.b64decode(handshake_key.encode())

        key_and_iv = keypair.decrypt_handshake_key(handshake_key_bytes)
        if key_and_iv is None:
            raise ValueError("Decryption failed!")

        return AesEncyptionSession(key_and_iv[:16], key_and_iv[16:])

    def __init__(self, key: bytes, iv: bytes) -> None:
        self.cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        self.padding_strategy = padding.PKCS7(algorithms.AES.block_size)

    def encrypt(self, data: bytes) -> bytes:
        """Encrypt the message."""
        encryptor = self.cipher.encryptor()
        padder = self.padding_strategy.padder()
        padded_data = padder.update(data) + padder.finalize()
        encrypted = encryptor.update(padded_data) + encryptor.finalize()
        return base64.b64encode(encrypted)

    def decrypt(self, data: str | bytes) -> str:
        """Decrypt the message."""
        decryptor = self.cipher.decryptor()
        unpadder = self.padding_strategy.unpadder()
        decrypted = decryptor.update(base64.b64decode(data)) + decryptor.finalize()
        unpadded_data = unpadder.update(decrypted) + unpadder.finalize()
        return unpadded_data.decode()


class KeyPair:
    """Class for generating key pairs."""

    @staticmethod
    def create_key_pair(key_size: int = 1024) -> KeyPair:
        """Create a key pair."""
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)
        public_key = private_key.public_key()
        return KeyPair(private_key, public_key)

    @staticmethod
    def create_from_der_keys(
        private_key_der_b64: str, public_key_der_b64: str
    ) -> KeyPair:
        """Create a key pair."""
        key_bytes = base64.b64decode(private_key_der_b64.encode())
        private_key = cast(
            rsa.RSAPrivateKey, serialization.load_der_private_key(key_bytes, None)
        )
        key_bytes = base64.b64decode(public_key_der_b64.encode())
        public_key = cast(
            rsa.RSAPublicKey, serialization.load_der_public_key(key_bytes, None)
        )

        return KeyPair(private_key, public_key)

    def __init__(
        self, private_key: rsa.RSAPrivateKey, public_key: rsa.RSAPublicKey
    ) -> None:
        self.private_key = private_key
        self.public_key = public_key
        self.private_key_der_bytes = self.private_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        self.public_key_der_bytes = self.public_key.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        self.private_key_der_b64 = base64.b64encode(self.private_key_der_bytes).decode()
        self.public_key_der_b64 = base64.b64encode(self.public_key_der_bytes).decode()

    def get_public_pem(self) -> bytes:
        """Get public key in PEM encoding."""
        return self.public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def decrypt_handshake_key(self, encrypted_key: bytes) -> bytes:
        """Decrypt an aes handshake key."""
        decrypted = self.private_key.decrypt(
            encrypted_key, asymmetric_padding.PKCS1v15()
        )
        return decrypted

    def decrypt_discovery_key(self, encrypted_key: bytes) -> bytes:
        """Decrypt an aes discovery key."""
        decrypted = self.private_key.decrypt(
            encrypted_key,
            asymmetric_padding.OAEP(
                mgf=asymmetric_padding.MGF1(algorithm=hashes.SHA1()),  # noqa: S303
                algorithm=hashes.SHA1(),  # noqa: S303
                label=None,
            ),
        )
        return decrypted



from __future__ import annotations

import re

import pytest
from voluptuous import (
    All,
    Boolean,
    Optional,
    Range,
    Schema,
)

from kasa import Device, IotLightPreset, KasaException, LightState, Module
from kasa.iot import IotBulb, IotDimmer
from kasa.iot.modules import LightPreset as IotLightPresetModule
from tests.conftest import (
    bulb_iot,
    color_bulb_iot,
    dimmable_iot,
    handle_turn_on,
    non_dimmable_iot,
    turn_on,
    variable_temp_iot,
)
from tests.iot.test_iotdevice import SYSINFO_SCHEMA


@bulb_iot
async def test_bulb_sysinfo(dev: Device):
    assert dev.sys_info is not None
    SYSINFO_SCHEMA_BULB(dev.sys_info)

    assert dev.model is not None


@bulb_iot
async def test_light_state_without_update(dev: IotBulb, monkeypatch):
    monkeypatch.setitem(dev._last_update["system"]["get_sysinfo"], "light_state", None)
    with pytest.raises(KasaException):
        print(dev.light_state)


@bulb_iot
async def test_get_light_state(dev: IotBulb):
    LIGHT_STATE_SCHEMA(await dev.get_light_state())


@color_bulb_iot
async def test_set_hsv_transition(dev: IotBulb, mocker):
    set_light_state = mocker.patch("kasa.iot.IotBulb._set_light_state")
    light = dev.modules.get(Module.Light)
    assert light
    await light.set_hsv(10, 10, 100, transition=1000)

    set_light_state.assert_called_with(
        {"hue": 10, "saturation": 10, "brightness": 100, "color_temp": 0},
        transition=1000,
    )


@bulb_iot
async def test_light_set_state(dev: IotBulb, mocker):
    """Testing setting LightState on the light module."""
    light = dev.modules.get(Module.Light)
    assert light
    set_light_state = mocker.spy(dev, "_set_light_state")
    state = LightState(light_on=True)
    await light.set_state(state)

    set_light_state.assert_called_with({"on_off": 1}, transition=None)
    state = LightState(light_on=False)
    await light.set_state(state)

    set_light_state.assert_called_with({"on_off": 0}, transition=None)


@variable_temp_iot
async def test_set_color_temp_transition(dev: IotBulb, mocker):
    set_light_state = mocker.patch("kasa.iot.IotBulb._set_light_state")
    light = dev.modules.get(Module.Light)
    assert light
    await light.set_color_temp(2700, transition=100)

    set_light_state.assert_called_with({"color_temp": 2700}, transition=100)


@variable_temp_iot
@pytest.mark.xdist_group(name="caplog")
async def test_unknown_temp_range(dev: IotBulb, monkeypatch, caplog):
    monkeypatch.setitem(dev._sys_info, "model", "unknown bulb")
    light = dev.modules.get(Module.Light)
    assert light
    color_temp_feat = light.get_feature("color_temp")
    assert color_temp_feat
    assert color_temp_feat.range == (2700, 5000)
    assert "Unknown color temperature range, fallback to 2700-5000" in caplog.text


@dimmable_iot
@turn_on
async def test_dimmable_brightness(dev: IotBulb, turn_on):
    assert isinstance(dev, IotBulb | IotDimmer)
    light = dev.modules.get(Module.Light)
    assert light
    await handle_turn_on(dev, turn_on)
    assert dev._is_dimmable

    await light.set_brightness(50)
    await dev.update()
    assert light.brightness == 50

    await light.set_brightness(10)
    await dev.update()
    assert light.brightness == 10

    with pytest.raises(TypeError, match="Brightness must be an integer"):
        await light.set_brightness("foo")  # type: ignore[arg-type]


@bulb_iot
async def test_turn_on_transition(dev: IotBulb, mocker):
    set_light_state = mocker.patch("kasa.iot.IotBulb._set_light_state")
    await dev.turn_on(transition=1000)

    set_light_state.assert_called_with({"on_off": 1}, transition=1000)

    await dev.turn_off(transition=100)

    set_light_state.assert_called_with({"on_off": 0}, transition=100)


@bulb_iot
async def test_dimmable_brightness_transition(dev: IotBulb, mocker):
    set_light_state = mocker.patch("kasa.iot.IotBulb._set_light_state")
    light = dev.modules.get(Module.Light)
    assert light
    await light.set_brightness(10, transition=1000)

    set_light_state.assert_called_with({"brightness": 10, "on_off": 1}, transition=1000)


@dimmable_iot
async def test_invalid_brightness(dev: IotBulb):
    assert dev._is_dimmable
    light = dev.modules.get(Module.Light)
    assert light
    with pytest.raises(
        ValueError,
        match=re.escape("Invalid brightness value: 110 (valid range: 0-100%)"),
    ):
        await light.set_brightness(110)

    with pytest.raises(
        ValueError,
        match=re.escape("Invalid brightness value: -100 (valid range: 0-100%)"),
    ):
        await light.set_brightness(-100)


@non_dimmable_iot
async def test_non_dimmable(dev: IotBulb):
    assert not dev._is_dimmable
    light = dev.modules.get(Module.Light)
    assert light
    with pytest.raises(KasaException):
        assert light.brightness == 0
    with pytest.raises(KasaException):
        await light.set_brightness(100)


@bulb_iot
async def test_ignore_default_not_set_without_color_mode_change_turn_on(
    dev: IotBulb, mocker
):
    query_helper = mocker.patch("kasa.iot.IotBulb._query_helper")
    # When turning back without settings, ignore default to restore the state
    await dev.turn_on()
    args, kwargs = query_helper.call_args_list[0]
    assert args[2] == {"on_off": 1, "ignore_default": 0}

    await dev.turn_off()
    args, kwargs = query_helper.call_args_list[1]
    assert args[2] == {"on_off": 0, "ignore_default": 1}


@bulb_iot
async def test_list_presets(dev: IotBulb):
    light_preset = dev.modules.get(Module.LightPreset)
    assert light_preset
    assert isinstance(light_preset, IotLightPresetModule)
    presets = light_preset._deprecated_presets
    # Light strip devices may list some light effects along with normal presets but these
    # are handled by the LightEffect module so exclude preferred states with id
    raw_presets = [
        pstate for pstate in dev.sys_info["preferred_state"] if "id" not in pstate
    ]
    assert len(presets) == len(raw_presets)

    for preset, raw in zip(presets, raw_presets, strict=False):
        assert preset.index == raw["index"]
        assert preset.brightness == raw["brightness"]
        assert preset.hue == raw["hue"]
        assert preset.saturation == raw["saturation"]
        assert preset.color_temp == raw["color_temp"]


@bulb_iot
async def test_modify_preset(dev: IotBulb, mocker):
    """Verify that modifying preset calls the and exceptions are raised properly."""
    if (
        not (light_preset := dev.modules.get(Module.LightPreset))
        or not light_preset._deprecated_presets
    ):
        pytest.skip("Some strips do not support presets")

    assert isinstance(light_preset, IotLightPresetModule)
    data: dict[str, int | None] = {
        "index": 0,
        "brightness": 10,
        "hue": 0,
        "saturation": 0,
        "color_temp": 0,
    }
    preset = IotLightPreset(**data)  # type: ignore[call-arg, arg-type]

    assert preset.index == 0
    assert preset.brightness == 10
    assert preset.hue == 0
    assert preset.saturation == 0
    assert preset.color_temp == 0

    await light_preset._deprecated_save_preset(preset)
    await dev.update()
    assert light_preset._deprecated_presets[0].brightness == 10

    with pytest.raises(KasaException):
        await light_preset._deprecated_save_preset(
            IotLightPreset(index=5, hue=0, brightness=0, saturation=0, color_temp=0)  # type: ignore[call-arg]
        )


@bulb_iot
@pytest.mark.parametrize(
    ("preset", "payload"),
    [
        (
            IotLightPreset(index=0, hue=0, brightness=1, saturation=0),  # type: ignore[call-arg]
            {"index": 0, "hue": 0, "brightness": 1, "saturation": 0},
        ),
        (
            IotLightPreset(index=0, brightness=1, id="testid", mode=2, custom=0),  # type: ignore[call-arg]
            {"index": 0, "brightness": 1, "id": "testid", "mode": 2, "custom": 0},
        ),
    ],
)
async def test_modify_preset_payloads(dev: IotBulb, preset, payload, mocker):
    """Test that modify preset payloads ignore none values."""
    if (
        not (light_preset := dev.modules.get(Module.LightPreset))
        or not light_preset._deprecated_presets
    ):
        pytest.skip("Some strips do not support presets")

    query_helper = mocker.patch("kasa.iot.IotBulb._query_helper")
    await light_preset._deprecated_save_preset(preset)
    query_helper.assert_called_with(dev.LIGHT_SERVICE, "set_preferred_state", payload)


LIGHT_STATE_SCHEMA = Schema(
    {
        "brightness": All(int, Range(min=0, max=100)),
        "color_temp": int,
        "hue": All(int, Range(min=0, max=360)),
        "mode": str,
        "on_off": Boolean,
        "saturation": All(int, Range(min=0, max=100)),
        "length": Optional(int),
        "transition": Optional(int),
        "dft_on_state": Optional(
            {
                "brightness": All(int, Range(min=0, max=100)),
                "color_temp": All(int, Range(min=0, max=9000)),
                "hue": All(int, Range(min=0, max=360)),
                "mode": str,
                "saturation": All(int, Range(min=0, max=100)),
                "groups": Optional(list[int]),
            }
        ),
        "err_code": int,
    }
)

SYSINFO_SCHEMA_BULB = SYSINFO_SCHEMA.extend(
    {
        "ctrl_protocols": Optional(dict),
        "description": Optional(str),  # Seen on LBxxx, similar to dev_name
        "dev_state": str,
        "disco_ver": str,
        "heapsize": int,
        "is_color": Boolean,
        "is_dimmable": Boolean,
        "is_factory": Boolean,
        "is_variable_color_temp": Boolean,
        "light_state": LIGHT_STATE_SCHEMA,
        "preferred_state": [
            {
                "brightness": All(int, Range(min=0, max=100)),
                "color_temp": int,
                "hue": All(int, Range(min=0, max=360)),
                "index": int,
                "saturation": All(int, Range(min=0, max=100)),
            }
        ],
    }
)


@bulb_iot
async def test_turn_on_behaviours(dev: IotBulb):
    behavior = await dev.get_turn_on_behavior()
    assert behavior



"""Module to mass update fixture files."""

import json
import logging
from collections.abc import Callable
from pathlib import Path

import asyncclick as click

from devtools.dump_devinfo import _wrap_redactors
from kasa.discover import NEW_DISCOVERY_REDACTORS, redact_data
from kasa.protocols.iotprotocol import REDACTORS as IOT_REDACTORS
from kasa.protocols.smartprotocol import REDACTORS as SMART_REDACTORS

FIXTURE_FOLDER = "tests/fixtures/"

_LOGGER = logging.getLogger(__name__)


def update_fixtures(update_func: Callable[[dict], bool], *, dry_run: bool) -> None:
    """Run the update function against the fixtures."""
    for file in Path(FIXTURE_FOLDER).glob("**/*.json"):
        with file.open("r") as f:
            fixture_data = json.load(f)

        if file.parent.name == "serialization":
            continue
        changed = update_func(fixture_data)
        if changed:
            click.echo(f"Will update {file.name}\n")
        if changed and not dry_run:
            with file.open("w") as f:
                json.dump(fixture_data, f, sort_keys=True, indent=4)
                f.write("\n")


def _discovery_result_update(info) -> bool:
    """Update discovery_result to be the raw result and error_code."""
    if (disco_result := info.get("discovery_result")) and "result" not in disco_result:
        info["discovery_result"] = {
            "result": disco_result,
            "error_code": 0,
        }
        return True
    return False


def _child_device_id_update(info) -> bool:
    """Update child device ids to be the scrubbed ids from dump_devinfo."""
    changed = False
    if get_child_device_list := info.get("get_child_device_list"):
        child_device_list = get_child_device_list["child_device_list"]
        child_component_list = info["get_child_device_component_list"][
            "child_component_list"
        ]
        for index, child_device in enumerate(child_device_list):
            child_component = child_component_list[index]
            if "SCRUBBED" not in child_device["device_id"]:
                dev_id = f"SCRUBBED_CHILD_DEVICE_ID_{index + 1}"
                click.echo(
                    f"child_device_id{index}: {child_device['device_id']} -> {dev_id}"
                )
                child_device["device_id"] = dev_id
                child_component["device_id"] = dev_id
                changed = True

    if children := info.get("system", {}).get("get_sysinfo", {}).get("children"):
        for index, child_device in enumerate(children):
            if "SCRUBBED" not in child_device["id"]:
                dev_id = f"SCRUBBED_CHILD_DEVICE_ID_{index + 1}"
                click.echo(f"child_device_id{index}: {child_device['id']} -> {dev_id}")
                child_device["id"] = dev_id
                changed = True

    return changed


def _diff_data(fullkey, data1, data2, diffs):
    if isinstance(data1, dict):
        for k, v in data1.items():
            _diff_data(fullkey + "/" + k, v, data2[k], diffs)
    elif isinstance(data1, list):
        for index, item in enumerate(data1):
            _diff_data(fullkey + "/" + str(index), item, data2[index], diffs)
    elif data1 != data2:
        diffs[fullkey] = (data1, data2)


def _redactor_result_update(info) -> bool:
    """Update fixtures with the output using the common redactors."""
    changed = False

    redactors = IOT_REDACTORS if "system" in info else SMART_REDACTORS

    for key, val in info.items():
        if not isinstance(val, dict):
            continue
        if key == "discovery_result":
            info[key] = redact_data(val, _wrap_redactors(NEW_DISCOVERY_REDACTORS))
        else:
            info[key] = redact_data(val, _wrap_redactors(redactors))
        diffs: dict[str, tuple[str, str]] = {}
        _diff_data(key, val, info[key], diffs)
        if diffs:
            for k, v in diffs.items():
                click.echo(f"{k}: {v[0]} -> {v[1]}")
            changed = True

    return changed


@click.option(
    "--dry-run/--no-dry-run",
    default=False,
    is_flag=True,
    type=bool,
    help="Perform a dry run without saving.",
)
@click.command()
async def cli(dry_run: bool) -> None:
    """Cli method fo rupdating fixtures."""
    update_fixtures(_discovery_result_update, dry_run=dry_run)
    update_fixtures(_child_device_id_update, dry_run=dry_run)
    update_fixtures(_redactor_result_update, dry_run=dry_run)


if __name__ == "__main__":
    cli()



"""Implementation of energy monitoring module."""

from __future__ import annotations

from typing import Any, NoReturn

from ...emeterstatus import EmeterStatus
from ...exceptions import DeviceError, KasaException
from ...interfaces.energy import Energy as EnergyInterface
from ..smartmodule import SmartModule, raise_if_update_error


class Energy(SmartModule, EnergyInterface):
    """Implementation of energy monitoring module."""

    REQUIRED_COMPONENT = "energy_monitoring"

    _energy: dict[str, Any]
    _current_consumption: float | None

    async def _post_update_hook(self) -> None:
        try:
            data = self.data
        except DeviceError as de:
            self._energy = {}
            self._current_consumption = None
            raise de

        # If version is 1 then data is get_energy_usage
        self._energy = data.get("get_energy_usage", data)

        if "voltage_mv" in data.get("get_emeter_data", {}):
            self._supported = (
                self._supported | EnergyInterface.ModuleFeature.VOLTAGE_CURRENT
            )

        if (power := self._energy.get("current_power")) is not None or (
            power := data.get("get_emeter_data", {}).get("power_mw")
        ) is not None:
            self._current_consumption = power / 1_000
        # Fallback if get_energy_usage does not provide current_power,
        # which can happen on some newer devices (e.g. P304M).
        # This may not be valid scenario as it pre-dates trying get_emeter_data
        elif (
            power := self.data.get("get_current_power", {}).get("current_power")
        ) is not None:
            self._current_consumption = power
        else:
            self._current_consumption = None

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        req = {
            "get_energy_usage": None,
        }
        if self.supported_version > 1:
            req["get_current_power"] = None
            req["get_emeter_data"] = None
            req["get_emeter_vgain_igain"] = None
        return req

    @property
    def optional_response_keys(self) -> list[str]:
        """Return optional response keys for the module."""
        if self.supported_version > 1:
            return ["get_energy_usage"]
        return []

    @property
    def current_consumption(self) -> float | None:
        """Current power in watts."""
        return self._current_consumption

    @property
    def energy(self) -> dict:
        """Return get_energy_usage results."""
        return self._energy

    def _get_status_from_energy(self, energy: dict) -> EmeterStatus:
        return EmeterStatus(
            {
                "power_mw": energy.get("current_power", 0),
                "total": energy.get("today_energy", 0) / 1_000,
            }
        )

    @property
    @raise_if_update_error
    def status(self) -> EmeterStatus:
        """Get the emeter status."""
        if "get_emeter_data" in self.data:
            return EmeterStatus(self.data["get_emeter_data"])
        else:
            return self._get_status_from_energy(self.energy)

    async def get_status(self) -> EmeterStatus:
        """Return real-time statistics."""
        if "get_emeter_data" in self.data:
            res = await self.call("get_emeter_data")
            return EmeterStatus(res["get_emeter_data"])
        else:
            res = await self.call("get_energy_usage")
            return self._get_status_from_energy(res["get_energy_usage"])

    @property
    def consumption_this_month(self) -> float | None:
        """Get the emeter value for this month in kWh."""
        if (month := self.energy.get("month_energy")) is not None:
            return month / 1_000
        return None

    @property
    def consumption_today(self) -> float | None:
        """Get the emeter value for today in kWh."""
        if (today := self.energy.get("today_energy")) is not None:
            return today / 1_000
        return None

    @property
    @raise_if_update_error
    def consumption_total(self) -> float | None:
        """Return total consumption since last reboot in kWh."""
        return None

    @property
    @raise_if_update_error
    def current(self) -> float | None:
        """Return the current in A."""
        if (ma := self.data.get("get_emeter_data", {}).get("current_ma")) is not None:
            return ma / 1_000
        return None

    @property
    @raise_if_update_error
    def voltage(self) -> float | None:
        """Get the current voltage in V."""
        if (mv := self.data.get("get_emeter_data", {}).get("voltage_mv")) is not None:
            return mv / 1_000
        return None

    async def _deprecated_get_realtime(self) -> EmeterStatus:
        """Retrieve current energy readings."""
        return self.status

    async def erase_stats(self) -> NoReturn:
        """Erase all stats."""
        raise KasaException("Device does not support periodic statistics")

    async def get_daily_stats(
        self, *, year: int | None = None, month: int | None = None, kwh: bool = True
    ) -> dict:
        """Return daily stats for the given year & month.

        The return value is a dictionary of {day: energy, ...}.
        """
        raise KasaException("Device does not support periodic statistics")

    async def get_monthly_stats(
        self, *, year: int | None = None, kwh: bool = True
    ) -> dict:
        """Return monthly stats for the given year."""
        raise KasaException("Device does not support periodic statistics")

    async def _check_supported(self) -> bool:
        """Additional check to see if the module is supported by the device."""
        # Energy module is not supported on P304M parent device
        return "device_on" in self._device.sys_info
